import pandas as pd
import pytest

from student_success_tool.preprocessing.features.pdp import term


@pytest.mark.parametrize(
    ["df", "first_term_of_year", "core_terms", "peak_covid_terms", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "student_guid": ["123", "123", "123", "123", "456", "456"],
                    "course_id": ["X101", "Y101", "X202", "Y202", "Z101", "Z202"],
                    "academic_year": [
                        "2020-21",
                        "2020-21",
                        "2020-21",
                        "2020-21",
                        "2019-20",
                        "2022-23",
                    ],
                    "academic_term": [
                        "FALL",
                        "FALL",
                        "WINTER",
                        "SPRING",
                        "FALL",
                        "SUMMER",
                    ],
                }
            ).astype(
                {
                    "academic_term": pd.CategoricalDtype(
                        ["FALL", "WINTER", "SPRING", "SUMMER"], ordered=True
                    )
                }
            ),
            "FALL",
            {"FALL", "SPRING"},
            {("2020-21", "SPRING"), ("2021-22", "FALL")},
            pd.DataFrame(
                {
                    "student_guid": ["123", "123", "123", "123", "456", "456"],
                    "course_id": ["X101", "Y101", "X202", "Y202", "Z101", "Z202"],
                    "academic_year": [
                        "2020-21",
                        "2020-21",
                        "2020-21",
                        "2020-21",
                        "2019-20",
                        "2022-23",
                    ],
                    "academic_term": [
                        "FALL",
                        "FALL",
                        "WINTER",
                        "SPRING",
                        "FALL",
                        "SUMMER",
                    ],
                    "term_id": [
                        "2020-21 FALL",
                        "2020-21 FALL",
                        "2020-21 WINTER",
                        "2020-21 SPRING",
                        "2019-20 FALL",
                        "2022-23 SUMMER",
                    ],
                    "term_start_dt": pd.to_datetime(
                        [
                            "2020-09-01",
                            "2020-09-01",
                            "2021-01-01",
                            "2021-02-01",
                            "2019-09-01",
                            "2023-06-01",
                        ],
                    ),
                    "term_rank": [1, 1, 2, 3, 0, 4],
                    "term_rank_core": [1, 1, pd.NA, 2, 0, pd.NA],
                    "term_rank_noncore": [pd.NA, pd.NA, 0, pd.NA, pd.NA, 1],
                    "term_in_peak_covid": [False, False, False, True, False, False],
                    "term_is_core": [True, True, False, True, True, False],
                    "term_is_noncore": [False, False, True, False, False, True],
                }
            ).astype(
                {
                    "term_rank_core": "Int8",
                    "term_rank_noncore": "Int8",
                    "academic_term": pd.CategoricalDtype(
                        ["FALL", "WINTER", "SPRING", "SUMMER"], ordered=True
                    ),
                }
            ),
        ),
    ],
)
def test_add_term_features(df, first_term_of_year, core_terms, peak_covid_terms, exp):
    obs = term.add_features(
        df,
        first_term_of_year=first_term_of_year,
        core_terms=core_terms,
        peak_covid_terms=peak_covid_terms,
    )
    assert isinstance(obs, pd.DataFrame)
    assert pd.testing.assert_frame_equal(obs, exp, check_dtype=False) is None


@pytest.mark.parametrize(
    ["df", "year_col", "term_col", "terms_subset", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "year": ["22-23", "23-24", "22-23", "21-22", "22-23", None],
                    "term": ["FA", "FA", "SP", "SP", "FA", None],
                }
            ).astype({"term": pd.CategoricalDtype(["FA", "SP"], ordered=True)}),
            "year",
            "term",
            None,
            pd.Series([1, 3, 2, 0, 1, pd.NA], dtype="Int8"),
        ),
    ],
)
def test_term_rank(df, year_col, term_col, terms_subset, exp):
    obs = term.term_rank(
        df, year_col=year_col, term_col=term_col, terms_subset=terms_subset
    )
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "year_col", "term_col", "peak_covid_terms", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "year": ["20-21", "20-21", "21-22", "21-22", "21-22"],
                    "term": ["FA", "SP", "FA", "SP", "SU"],
                }
            ),
            "year",
            "term",
            [("20-21", "SP"), ("20-21", "SU"), ("21-22", "FA")],
            pd.Series([False, True, True, False, False], dtype="bool"),
        )
    ],
)
def test_term_in_peak_covid(df, year_col, term_col, peak_covid_terms, exp):
    obs = term.term_in_peak_covid(
        df, year_col=year_col, term_col=term_col, peak_covid_terms=peak_covid_terms
    )
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "terms_subset", "term_col", "exp"],
    [
        (
            pd.DataFrame({"term": ["FALL", "WINTER", "SPRING", "SUMMER"]}),
            {"FALL", "SPRING"},
            "term",
            pd.Series([True, False, True, False], dtype="boolean"),
        ),
        (
            pd.DataFrame({"term": ["FALL", "WINTER", "SPRING", "SUMMER"]}),
            {"FALL", "WINTER", "SPRING"},
            "term",
            pd.Series([True, True, True, False], dtype="boolean"),
        ),
    ],
)
def test_term_in_subset(df, terms_subset, term_col, exp):
    obs = term.term_in_subset(df, terms_subset=terms_subset, term_col=term_col)
    assert isinstance(obs, pd.Series)
    assert pd.testing.assert_series_equal(obs, exp, check_names=False) is None



import pandas as pd
import pytest

from student_success_tool.preprocessing.features.pdp import student_term

# @pytest.mark.parametrize(
#     ["df", "exp"],
#     [
#         (
#             pd.DataFrame(
#                 {
#                     "cohort": ["20-21", "20-21", "20-21", "22-23", "22-23"],
#                     "academic_year": ["20-21", "21-22", "23-24", "22-23", "23-24"],
#                     "num_credits_earned": [5.0, 10.0, 6.0, 0.0, 15.0],
#                     "num_credits_attempted": [10.0, 10.0, 8.0, 15.0, 15.0],
#                 }
#             ),
#             pd.DataFrame(
#                 {
#                     "year_of_enrollment_at_cohort_inst": [1, 2, 4, 1, 2],
#                     "frac_credits_earned": [0.5, 1.0, 0.75, 0.0, 1.0],
#                 }
#             ),
#         ),
#     ],
# )
# def test_add_student_term_features(df, exp):
#     obs = student_term.add_features(df)
#     assert isinstance(obs, pd.DataFrame) and not obs.empty
#     assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "grp_cols", "agg_cols", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "student_guid": ["123", "123", "123", "456", "456", "789"],
                    "term_id": [
                        "23-24 FALL",
                        "23-24 FALL",
                        "23-24 SPRING",
                        "23-24 FALL",
                        "23-24 FALL",
                        "23-24 SPRING",
                    ],
                    "course_type": ["CU", "CD", "CU", "CU", "CC", "CU"],
                    "course_level": [1, 0, 1, 2, 0, 1],
                }
            ),
            ["student_guid", "term_id"],
            ["course_type", "course_level"],
            pd.DataFrame(
                {
                    "student_guid": ["123", "123", "456", "789"],
                    "term_id": [
                        "23-24 FALL",
                        "23-24 SPRING",
                        "23-24 FALL",
                        "23-24 SPRING",
                    ],
                    "num_courses_course_type_CC": [0, 0, 1, 0],
                    "num_courses_course_type_CD": [1, 0, 0, 0],
                    "num_courses_course_type_CU": [1, 1, 1, 1],
                    "num_courses_course_level_0": [1, 0, 1, 0],
                    "num_courses_course_level_1": [1, 1, 0, 1],
                    "num_courses_course_level_2": [0, 0, 1, 0],
                }
            ),
        ),
    ],
)
def test_sum_dummy_cols_by_group(df, grp_cols, agg_cols, exp):
    obs = student_term.sum_dummy_cols_by_group(df, grp_cols=grp_cols, agg_cols=agg_cols)
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "grp_cols", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "student_guid": ["123", "123", "456", "789"],
                    "term_id": [
                        "23-24 FALL",
                        "23-24 SPRING",
                        "23-24 FALL",
                        "23-24 SPRING",
                    ],
                    "num_courses_course_type_CC|CD": [0, 3, 1, 0],
                    "num_courses_course_id_eng_101": [0, 0, 0, 1],
                    "num_courses_course_subject_area_51": [2, 1, 1, 1],
                }
            ),
            ["student_guid", "term_id"],
            pd.DataFrame(
                {
                    "student_guid": ["123", "123", "456", "789"],
                    "term_id": [
                        "23-24 FALL",
                        "23-24 SPRING",
                        "23-24 FALL",
                        "23-24 SPRING",
                    ],
                    "took_course_id_eng_101": [False, False, False, True],
                    "took_course_subject_area_51": [True, True, True, True],
                }
            ),
        ),
    ],
)
def test_equal_cols_by_group(df, grp_cols, exp):
    obs = student_term.equal_cols_by_group(df, grp_cols=grp_cols)
    print("obs columns", obs.columns)
    print("exp cols", exp.columns)
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "grp_cols", "agg_col_vals", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "student_guid": ["123", "123", "123", "456", "456", "789"],
                    "term_id": [
                        "23-24 FALL",
                        "23-24 FALL",
                        "23-24 SPRING",
                        "23-24 FALL",
                        "23-24 FALL",
                        "23-24 SPRING",
                    ],
                    "course_type": ["CU", "CD", "CU", "CU", "CC", "CU"],
                    "course_level": [1, 0, 1, 2, 0, 1],
                    "grade": ["F", "F", "P", "W", "F", "P"],
                }
            ),
            ["student_guid", "term_id"],
            [
                ("course_type", ["CC", "CD"]),
                ("course_level", 0),
                ("course_level", [2, 3]),
                ("grade", ["0", "1", "F", "W"]),
            ],
            pd.DataFrame(
                {
                    "student_guid": ["123", "123", "456", "789"],
                    "term_id": [
                        "23-24 FALL",
                        "23-24 SPRING",
                        "23-24 FALL",
                        "23-24 SPRING",
                    ],
                    "num_courses_course_type_CC|CD": [1, 0, 1, 0],
                    "num_courses_course_level_0": [1, 0, 1, 0],
                    "num_courses_course_level_2|3": [0, 0, 1, 0],
                    "num_courses_grade_0|1|F|W": [2, 0, 2, 0],
                }
            ),
        ),
    ],
)
def test_sum_val_equal_cols_by_group(df, grp_cols, agg_col_vals, exp):
    obs = student_term.sum_val_equal_cols_by_group(
        df, grp_cols=grp_cols, agg_col_vals=agg_col_vals
    )
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    [
        "df",
        "min_passing_grade",
        "grp_cols",
        "grade_col",
        "grade_numeric_col",
        "section_grade_numeric_col",
        "exp",
    ],
    [
        (
            pd.DataFrame(
                {
                    "sid": ["123", "123", "123", "123", "456", "456"],
                    "tid": [
                        "22-23 FA",
                        "22-23 FA",
                        "22-23 FA",
                        "22-23 SP",
                        "22-23 SP",
                        "22-23 SP",
                    ],
                    "grade": ["4", "3", "F", "1", pd.NA, "4"],
                    "grade_num": [4.0, 3.0, pd.NA, 1.0, pd.NA, 4.0],
                    "section_grade_num_mean": [3.25, 3.0, 2.75, 2.5, 3.0, 3.5],
                }
            ).astype({"grade": "string", "grade_num": "Float32"}),
            1.0,
            ["sid", "tid"],
            "grade",
            "grade_num",
            "section_grade_num_mean",
            pd.DataFrame(
                {
                    "sid": ["123", "123", "456"],
                    "tid": ["22-23 FA", "22-23 SP", "22-23 SP"],
                    "num_courses_grade_is_failing_or_withdrawal": [1, 0, 0],
                    "num_courses_grade_above_section_avg": [1, 0, 1],
                }
            ),
        ),
    ],
)
def test_multicol_grade_aggs_by_group(
    df,
    min_passing_grade,
    grp_cols,
    grade_col,
    grade_numeric_col,
    section_grade_numeric_col,
    exp,
):
    obs = student_term.multicol_grade_aggs_by_group(
        df,
        min_passing_grade=min_passing_grade,
        grp_cols=grp_cols,
        grade_col=grade_col,
        grade_numeric_col=grade_numeric_col,
        section_grade_numeric_col=section_grade_numeric_col,
    )
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "ccol", "tcol", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "cohort_start_dt": ["2019-09-01", "2019-09-01", "2021-02-01"],
                    "term_start_dt": ["2020-02-01", "2020-09-01", "2023-09-01"],
                },
                dtype="datetime64[s]",
            ),
            "cohort_start_dt",
            "term_start_dt",
            pd.Series([1, 2, 3], dtype="Int8"),
        ),
    ],
)
def test_year_of_enrollment_at_cohort_inst(df, ccol, tcol, exp):
    obs = student_term.year_of_enrollment_at_cohort_inst(
        df, cohort_start_dt_col=ccol, term_start_dt_col=tcol
    )
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "inst", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "year_of_enrollment_at_cohort_inst": [1, 2, 3, 4],
                    "first_year_to_certificate_at_cohort_inst": [
                        2,
                        pd.NA,
                        2,
                        2,
                    ],
                    "years_to_latest_certificate_at_cohort_inst": [
                        3,
                        3,
                        pd.NA,
                        3,
                    ],
                    "first_year_to_certificate_at_other_inst": [
                        2,
                        pd.NA,
                        2,
                        2,
                    ],
                    "years_to_latest_certificate_at_other_inst": [
                        3,
                        3,
                        pd.NA,
                        3,
                    ],
                },
                dtype="Int8",
            ),
            "cohort",
            pd.Series([False, False, True, True], dtype="boolean"),
        ),
    ],
)
def test_student_earned_certificate(df, inst, exp):
    obs = student_term.student_earned_certificate(df, inst=inst)
    assert isinstance(obs, pd.Series) and not obs.empty
    assert pd.testing.assert_series_equal(obs, exp) is None


@pytest.mark.parametrize(
    ["df", "ccol", "tcol", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "cohort_start_dt": ["2019-09-01", "2019-09-01", "2021-02-01"],
                    "term_start_dt": ["2020-02-01", "2019-09-01", "2020-09-01"],
                },
                dtype="datetime64[s]",
            ),
            "cohort_start_dt",
            "term_start_dt",
            pd.Series([False, False, True], dtype="boolean"),
        ),
    ],
)
def test_term_is_pre_cohort(df, ccol, tcol, exp):
    obs = student_term.term_is_pre_cohort(
        df, cohort_start_dt_col=ccol, term_start_dt_col=tcol
    )
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "study_area_col", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "study_area_term_1": ["01", "02", None, "03"],
                    "study_area_year_1": ["01", "03", None, "03"],
                    "course_subject_areas": [
                        ["01", "01", "01", "02"],
                        ["01", "02", "01"],
                        ["01", "02", "03"],
                        [],
                    ],
                }
            ).astype({"study_area_term_1": "string", "study_area_year_1": "string"}),
            "study_area_term_1",
            pd.Series([3, 1, 0, 0], dtype="Int8"),
        ),
        (
            pd.DataFrame(
                {
                    "study_area_term_1": ["01", "02", None, "03"],
                    "study_area_year_1": ["01", "03", None, "03"],
                    "course_subject_areas": [
                        ["01", "01", "01", "02"],
                        ["01", "02", "01"],
                        ["01", "02", "03"],
                        [],
                    ],
                }
            ).astype({"study_area_term_1": "string", "study_area_year_1": "string"}),
            "study_area_year_1",
            pd.Series([3, 0, 0, 0], dtype="Int8"),
        ),
    ],
)
def test_num_courses_in_study_area(df, study_area_col, exp):
    obs = student_term.num_courses_in_study_area(df, study_area_col=study_area_col)
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "numer_col", "denom_col", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "num_courses": [1, 2, 4, 5, 7],
                    "num_courses_passed": [1, 1, 3, 4, 0],
                }
            ),
            "num_courses_passed",
            "num_courses",
            pd.Series([1.0, 0.5, 0.75, 0.8, 0.0]),
        ),
    ],
)
def test_compute_frac_courses(df, numer_col, denom_col, exp):
    obs = student_term.compute_frac_courses(
        df, numer_col=numer_col, denom_col=denom_col
    )
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "student_col", "sections_col", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "frac_courses_passed": [1.0, 0.5, 0.75],
                    "frac_sections_students_passed": [0.9, 0.5, 0.8],
                }
            ),
            "frac_courses_passed",
            "frac_sections_students_passed",
            pd.Series([True, False, False]),
        ),
    ],
)
def test_student_rate_above_sections_avg(df, student_col, sections_col, exp):
    obs = student_term.student_rate_above_sections_avg(
        df, student_col=student_col, sections_col=sections_col
    )
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "min_num_credits_full_time", "num_credits_col", "exp"],
    [
        (
            pd.DataFrame({"num_credits_attempted": [15.0, 12.0, 8.0, 0.0]}),
            12.0,
            "num_credits_attempted",
            pd.Series(["FULL-TIME", "FULL-TIME", "PART-TIME", "PART-TIME"]),
        ),
    ],
)
def test_student_term_enrollment_intensity(
    df, min_num_credits_full_time, num_credits_col, exp
):
    obs = student_term.student_term_enrollment_intensity(
        df,
        min_num_credits_full_time=min_num_credits_full_time,
        num_credits_col=num_credits_col,
    )
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty



try:
    import tomllib  # noqa
except ImportError:  # => PY3.10
    import tomli as tomllib  # noqa

import pydantic as pyd
import pathlib
import pytest

from student_success_tool.configs import custom


SRC_ROOT = pathlib.Path(__file__).parents[2] / "pipelines" / "custom" / "institution_id"


@pytest.fixture(scope="module")
def template_cfg_dict():
    config_path = SRC_ROOT / "config-TEMPLATE.toml"
    with config_path.open("rb") as f:
        return tomllib.load(f)


def test_template_pdp_cfgs(template_cfg_dict):
    result = custom.CustomProjectConfig.model_validate(template_cfg_dict)
    print(result)
    assert isinstance(result, pyd.BaseModel)


@pytest.mark.parametrize(
    ["cfg_str", "context"],
    [
        (
            'institution_id = "custom_inst_id"',
            pytest.raises(pyd.ValidationError),
        ),
        (
            """
            institution_id = "INVALID_IDENTIFIER!"
            institution_name = "Custom Institution Name"
            """,
            pytest.raises(pyd.ValidationError),
        ),
        (
            """
            institution_id = "custom_inst_id"
            institution_name = "Custom Institution Name"
            [datasets.bronze]

            [datasets.bronze.raw_cohort]
            primary_keys = ["student_id"]
            non_null_cols = ["acad_year"]
            train_file_path = "/Volumes/CATALOG/INST_ID_bronze/.../FILE_NAME_cohort_train.csv"
            predict_file_path = "/Volumes/CATALOG/INST_ID_bronze/.../FILE_NAME_cohort_inference.csv"
            """,
            pytest.raises(pyd.ValidationError),
        ),
        (
            """
            institution_id = "inst_id"
            institution_name = "Inst Name"

            [model]
            experiment_id = "EXPERIMENT_ID"
            run_id = "RUN_ID"
            framework = "sklearn"
            """,
            pytest.raises(pyd.ValidationError),
        ),
    ],
)
def test_bad_custom_cfgs(cfg_str, context):
    cfg = tomllib.loads(cfg_str)
    with context:
        result = custom.CustomProjectConfig.model_validate(cfg)
        assert isinstance(result, pyd.BaseModel)



import pytest
from unittest.mock import patch, MagicMock
import pandas as pd

from student_success_tool.reporting.model_card.base import ModelCard


@pytest.fixture
def mock_config():
    cfg = MagicMock()
    cfg.model = MagicMock(
        mlflow_model_uri="uri", framework="sklearn", run_id="123", experiment_id="456"
    )
    cfg.institution_id = "inst"
    cfg.institution_name = "TestInstitution"
    cfg.modeling.feature_selection.collinear_threshold = 0.9
    cfg.modeling.feature_selection.low_variance_threshold = 0.01
    cfg.modeling.feature_selection.incomplete_threshold = 0.05
    cfg.split_col = None
    return cfg


@pytest.fixture
def mock_client():
    return MagicMock()


def test_init_defaults(mock_config, mock_client):
    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    assert card.model_name == "inst_my_model"
    assert card.uc_model_name == "catalog.inst_gold.inst_my_model"
    assert card.assets_folder == "card_assets"
    assert isinstance(card.context, dict)


@patch("student_success_tool.reporting.model_card.base.dataio.models.load_mlflow_model")
def test_load_model_success(mock_load_model, mock_config, mock_client):
    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    card.load_model()
    mock_load_model.assert_called_once_with("uri", "sklearn")
    assert card.run_id == "123"
    assert card.experiment_id == "456"


def test_find_model_version_found(mock_config, mock_client):
    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    card.run_id = "123"
    mock_version = MagicMock(run_id="123", version="5")
    mock_client.search_model_versions.return_value = [mock_version]
    card.find_model_version()
    assert card.context["version_number"] == "5"


def test_find_model_version_not_found(mock_config, mock_client):
    mock_client.search_model_versions.return_value = []
    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    card.run_id = "999"
    card.find_model_version()
    assert card.context["version_number"] is None


def test_get_feature_metadata_success(mock_config, mock_client):
    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    card.model = MagicMock()
    card.model.named_steps = {
        "column_selector": MagicMock(get_params=lambda: {"cols": ["a", "b", "c"]})
    }
    metadata = card.get_feature_metadata()
    assert metadata["number_of_features"] == "3"
    assert metadata["collinearity_threshold"] == "0.9"


@patch("student_success_tool.reporting.model_card.base.utils.download_static_asset")
def test_get_basic_context(mock_download, mock_config, mock_client):
    mock_download.return_value = "<img>Logo</img>"
    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    result = card.get_basic_context()
    assert result["institution_name"] == "TestInstitution"
    assert "logo" in result


def test_build_calls_all_steps(mock_config, mock_client):
    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    for method in [
        "load_model",
        "find_model_version",
        "extract_training_data",
        "_register_sections",
        "collect_metadata",
        "render",
    ]:
        setattr(card, method, MagicMock())

    card.build()

    for method in [
        card.load_model,
        card.find_model_version,
        card.extract_training_data,
        card._register_sections,
        card.collect_metadata,
        card.render,
    ]:
        method.assert_called_once()


@patch("student_success_tool.reporting.model_card.base.utils.safe_count_runs")
@patch(
    "student_success_tool.reporting.model_card.base.modeling.evaluation.extract_training_data_from_model"
)
@patch("student_success_tool.reporting.model_card.base.dataio.models.load_mlflow_model")
def test_extract_training_data_with_split_call_load_model(
    mock_load_model, mock_extract_data, mock_safe_count, mock_config, mock_client
):
    mock_config.split_col = "split"
    df = pd.DataFrame(
        {"feature": [1, 2, 3, 4], "split": ["train", "test", "train", "val"]}
    )
    mock_extract_data.return_value = df
    mock_safe_count.return_value = 2  # simulate 2 runs in experiment

    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    card.load_model()
    card.extract_training_data()

    assert card.context["training_dataset_size"] == 2
    assert card.context["num_runs_in_experiment"] == 2


@patch("builtins.open", new_callable=MagicMock)
def test_render_template_and_output(mock_open, mock_config, mock_client):
    mock_file = MagicMock()
    mock_open.return_value.__enter__.return_value = mock_file
    mock_file.read.return_value = "Model: {institution_name}"

    card = ModelCard(
        config=mock_config,
        catalog="catalog",
        model_name="inst_my_model",
        mlflow_client=mock_client,
    )
    card.template_path = "template.md"
    card.output_path = "output.md"
    card.context = {"institution_name": "TestInstitution"}
    card.render()

    mock_open.assert_any_call("template.md", "r")
    mock_open.assert_any_call("output.md", "w")
    mock_file.write.assert_called_once_with("Model: TestInstitution")



import os
import logging
import typing as t

from mlflow.tracking import MlflowClient

# export .md to .pdf
import markdown
from weasyprint import HTML

# resolving files in templates module within package
from importlib.abc import Traversable
from importlib.resources import files

# internal SST modules
from ... import dataio, modeling

# relative imports in 'reporting' module
from ..sections import register_sections
from ..sections.registry import SectionRegistry
from ..utils import utils
from ..utils.formatting import Formatting
from ..utils.types import ModelCardConfig

LOGGER = logging.getLogger(__name__)
C = t.TypeVar("C", bound=ModelCardConfig)


class ModelCard(t.Generic[C]):
    DEFAULT_ASSETS_FOLDER = "card_assets"
    TEMPLATE_FILENAME = "model-card-TEMPLATE.md"
    LOGO_FILENAME = "logo.png"

    def __init__(
        self,
        config: C,
        catalog: str,
        model_name: str,
        assets_path: t.Optional[str] = None,
        mlflow_client: t.Optional[MlflowClient] = None,
    ):
        """
        Initializes the ModelCard object with the given config and the model name
        in unity catalog. If assets_path is not provided, the default assets folder is used.
        """
        self.cfg = config
        self.catalog = catalog
        self.model_name = model_name
        self.uc_model_name = f"{catalog}.{self.cfg.institution_id}_gold.{model_name}"
        LOGGER.info("Initializing ModelCard for model: %s", self.uc_model_name)

        self.client = mlflow_client or MlflowClient()
        self.section_registry = SectionRegistry()
        self.format = Formatting()
        self.context: dict[str, t.Any] = {}

        self.assets_folder = assets_path or self.DEFAULT_ASSETS_FOLDER
        self.output_path = self._build_output_path()
        self.template_path = self._resolve(
            "student_success_tool.reporting.template", self.TEMPLATE_FILENAME
        )
        self.logo_path = self._resolve(
            "student_success_tool.reporting.template.assets", self.LOGO_FILENAME
        )

    def build(self):
        """
        Builds the model card by performing the following steps:
        1. Loads the MLflow model.
        2. Finds the model version from the MLflow client based on the run ID.
        3. Extracts the training data from the MLflow run.
        4. Registers all sections in the section registry.
        5. Collects all metadata for the model card.
        6. Renders the model card using the template and context.
        """
        self.load_model()
        self.find_model_version()
        self.extract_training_data()
        self._register_sections()
        self.collect_metadata()
        self.render()

    def load_model(self):
        """
        Loads the MLflow model from the MLflow client based on the MLflow model URI.
        Also assigns the run ID and experiment ID from the config.
        """
        model_cfg = self.cfg.model
        if not model_cfg:
            raise ValueError(f"Model configuration for '{self.model_name}' is missing.")
        if not all(
            [model_cfg.mlflow_model_uri, model_cfg.run_id, model_cfg.experiment_id]
        ):
            raise ValueError(
                f"Incomplete model config for '{self.model_name}': "
                f"URI, run_id, or experiment_id missing."
            )

        self.model = dataio.models.load_mlflow_model(
            model_cfg.mlflow_model_uri,
            model_cfg.framework,
        )
        self.run_id = model_cfg.run_id
        self.experiment_id = model_cfg.experiment_id

    def find_model_version(self):
        """
        Retrieves the model version from the MLflow client based on the run ID.
        """
        try:
            versions = self.client.search_model_versions(f"name='{self.uc_model_name}'")
            for v in versions:
                if v.run_id == self.run_id:
                    self.context["version_number"] = v.version
                    LOGGER.info(f"Model Version = {self.context['version_number']}")
                    return
            LOGGER.warning(f"Unable to find model version for run id: {self.run_id}")
            self.context["version_number"] = None
        except Exception as e:
            LOGGER.error(
                f"Error retrieving model version for run id {self.run_id}: {e}"
            )
            self.context["version_number"] = None

    def extract_training_data(self):
        """
        Extracts the training data from the MLflow run utilizing SST internal subpackages (modeling).
        """
        self.modeling_data = modeling.evaluation.extract_training_data_from_model(
            self.experiment_id
        )
        self.training_data = self.modeling_data
        if self.cfg.split_col:
            self.training_data = self.modeling_data[
                self.modeling_data[self.cfg.split_col] == "train"
            ]
        self.context["training_dataset_size"] = self.training_data.shape[0]
        self.context["num_runs_in_experiment"] = utils.safe_count_runs(
            self.experiment_id
        )

    def collect_metadata(self):
        """
        Gathers all metadata for the model card. All of this data is dynamic and will
        depend on the institution and model. This calls functions that retrieves & downloads
        mlflow artifacts and also retrieves config information.
        """
        metadata_functions = [
            self.get_basic_context,
            self.get_feature_metadata,
            self.get_model_plots,
            self.section_registry.render_all,
        ]

        for func in metadata_functions:
            LOGGER.info(f"Updating context from {func.__name__}()")
            self.context.update(func())

    def get_basic_context(self) -> dict[str, str]:
        """
        Collects "basic" context which instantiates the DataKind logo, the
        institution name, and the current year.

        Returns:
            A dictionary with the keys as the variable names that will be called
            dynamically in template with values for each variable.
        """
        return {
            "logo": utils.download_static_asset(
                description="Logo",
                static_path=self.logo_path,
                local_folder=self.assets_folder,
            )
            or "",
            "institution_name": self.cfg.institution_name,
        }

    def get_feature_metadata(self) -> dict[str, str]:
        """
        Collects feature count from the MLflow run. Also, collects feature selection data
        from the config file.

        Returns:
            A dictionary with the keys as the variable names that will be called
            dynamically in template with values for each variable.
        """
        feature_count = len(
            self.model.named_steps["column_selector"].get_params()["cols"]
        )
        if not self.cfg.modeling or not self.cfg.modeling.feature_selection:
            raise ValueError(
                "Modeling configuration or feature selection config is missing."
            )

        fs_cfg = self.cfg.modeling.feature_selection

        return {
            "number_of_features": str(feature_count),
            "collinearity_threshold": str(fs_cfg.collinear_threshold),
            "low_variance_threshold": str(fs_cfg.low_variance_threshold),
            "incomplete_threshold": str(fs_cfg.incomplete_threshold),
        }

    def get_model_plots(self) -> dict[str, str]:
        """
        Collects model plots from the MLflow run, downloads them locally. These will later be
        rendered in the template.

        Returns:
            A dictionary with the keys as the plot names called in the template
            and the values are inline HTML (since these are all images) for each
            of the artifacts.
        """
        plots = {
            "model_comparison_plot": (
                "Model Comparison",
                "model_comparison.png",
                "125mm",
            ),
            "test_calibration_curve": (
                "Test Calibration Curve",
                "calibration/test_calibration.png",
                "125mm",
            ),
            "test_roc_curve": (
                "Test ROC Curve",
                "test_roc_curve_plot.png",
                "125mm",
            ),
            "test_confusion_matrix": (
                "Test Confusion Matrix",
                "test_confusion_matrix.png",
                "125mm",
            ),
            "test_histogram": (
                "Test Histogram",
                "preds/test_hist.png",
                "125mm",
            ),
            "feature_importances_by_shap_plot": (
                "Feature Importances",
                "feature_importances_by_shap_plot.png",
                "150mm",
            ),
        }
        return {
            key: utils.download_artifact(
                run_id=self.run_id,
                description=description,
                artifact_path=path,
                local_folder=self.assets_folder,
                fixed_width=width,
            )
            or ""
            for key, (description, path, width) in plots.items()
        }

    def render(self):
        """
        Renders the model card using the template and context data.
        """
        with open(self.template_path, "r") as file:
            template = file.read()
        filled = template.format(**self.context)
        with open(self.output_path, "w") as file:
            file.write(filled)
        LOGGER.info(f"✅ Model card generated at {self.output_path}")

    def reload_card(self):
        """
        Reloads Markdown model card post user editing after rendering.
        This offers flexibility in case user wants to utilize this class
        as a base and then makes edits in markdown before exporting as a PDF.
        """
        # Read the Markdown output
        with open(self.output_path, "r") as f:
            self.md_content = f.read()
        LOGGER.info("Reloaded model card content")

    def style_card(self):
        """
        Styles card using CSS.
        """
        # Convert Markdown to HTML
        html_content = markdown.markdown(
            self.md_content,
            extensions=["extra", "tables", "sane_lists", "toc", "smarty"],
        )

        # Load CSS from external file
        css_path = self._resolve(
            "student_success_tool.reporting.template.styles", "model_card.css"
        )
        with open(css_path, "r") as f:
            style = f"<style>\n{f.read()}\n</style>"

        # Prepend CSS to HTML
        self.html_content = style + html_content
        LOGGER.info("Applied CSS styling")

    def export_to_pdf(self):
        """
        Export CSS styled HTML to PDF utilizing weasyprint for conversion.
        Also logs the card, so it can be accessed as a PDF in the run artifacts.
        """
        # Styles card using model_card.css
        self.style_card()

        # Define PDF path
        base_path = os.path.dirname(self.output_path) or "."
        self.pdf_path = self.output_path.replace(".md", ".pdf")

        # Render PDF
        try:
            HTML(string=self.html_content, base_url=base_path).write_pdf(self.pdf_path)
            LOGGER.info(f"✅ PDF model card saved to {self.pdf_path}")
        except Exception as e:
            raise RuntimeError(f"Failed to create PDF: {e}")

        # Save model card into gold volume for access web app compatibility
        utils.save_card_to_gold_volume(
            filename=self.pdf_path,
            catalog=self.catalog,
            institution_id=self.cfg.institution_id,
        )

    def _build_output_path(self) -> str:
        """
        Builds the output path for the model card.
        """
        filename = f"model-card-{self.model_name}.md"
        return os.path.join(os.getcwd(), filename)

    def _register_sections(self):
        """
        Registers all sections in the section registry.
        """
        register_sections(self, self.section_registry)

    def _resolve(self, package: str, filename: str) -> Traversable:
        """
        Resolves files using importlib. Importlib is necessary since
        the file exists within the SST package itself.
        """
        return files(package).joinpath(filename)



# ruff: noqa: F821
# mypy: ignore-errors
import functools as ft
import logging
import typing as t

import pandas as pd

try:
    import pandera as pda
    import pandera.typing as pt
except ModuleNotFoundError:
    from ... import utils

    utils.databricks.mock_pandera()

    import pandera as pda
    import pandera.typing as pt

LOGGER = logging.getLogger(__name__)

TermField = ft.partial(
    pda.Field,
    dtype_kwargs={
        "categories": ["FALL", "WINTER", "SPRING", "SUMMER"],
        "ordered": True,
    },
)
GPAField = ft.partial(pda.Field, nullable=True, ge=0.0, le=4.0)
NumCreditsGt0Field = ft.partial(pda.Field, nullable=True, ge=0.0)
GradeField = ft.partial(pda.Field, nullable=True)
CompletedGatewayField = ft.partial(
    pda.Field, nullable=True, dtype_kwargs={"categories": ["C", "D", "NA"]}
)
CompletedDevField = ft.partial(
    pda.Field, nullable=True, dtype_kwargs={"categories": ["Y", "N", "NA"]}
)
YearsToOfField = ft.partial(pda.Field, ge=0, le=8)


class RawPDPCohortDataSchema(pda.DataFrameModel):
    """
    Schema (aka ``DataFrameModel``) for raw PDP cohort data that validates columns,
    data types (including categorical categories), acceptable value ranges, and more.

    References:
        - https://help.studentclearinghouse.org/pdp/knowledge-base/cohort-level-analysis-ready-file-data-dictionary/
        - https://pandera.readthedocs.io/en/stable/dataframe_models.html
    """

    student_id: pt.Series["string"]
    institution_id: pt.Series["string"]
    cohort: pt.Series["string"]
    cohort_term: pt.Series[pd.CategoricalDtype] = TermField()
    enrollment_type: pt.Series[pd.CategoricalDtype]
    enrollment_intensity_first_term: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True
    )
    # NOTE: categories set in a parser, which forces "UK" values to null
    math_placement: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    english_placement: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    dual_and_summer_enrollment: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True
    )
    student_age: pt.Series[pd.StringDtype] = pda.Field(nullable=True)
    race: pt.Series[pd.StringDtype] = pda.Field(nullable=True)
    ethnicity: pt.Series[pd.StringDtype] = pda.Field(nullable=True)
    gender: pt.Series[pd.StringDtype] = pda.Field(nullable=True)
    first_gen: pt.Series[pd.StringDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    pell_status_first_year: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    attendance_status_term_1: pt.Series[pd.StringDtype] = pda.Field(nullable=True)
    credential_type_sought_year_1: pt.Series[pd.StringDtype] = pda.Field(nullable=True)
    program_of_study_term_1: pt.Series["string"] = pda.Field(nullable=True)
    gpa_group_term_1: pt.Series["Float32"] = GPAField()
    gpa_group_year_1: pt.Series["Float32"] = GPAField()
    number_of_credits_attempted_year_1: pt.Series["Float32"] = pda.Field(
        nullable=True, ge=1.0, raise_warning=True
    )
    number_of_credits_earned_year_1: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_2: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_2: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_3: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_3: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_4: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_4: pt.Series["Float32"] = NumCreditsGt0Field()
    # NOTE: categories set in a parser, which forces "UK" values to null
    gateway_math_status: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    gateway_english_status: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    attempted_gateway_math_year_1: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True
    )
    # NOTE: categories set in a parser, which forces "UK" values to null
    attempted_gateway_english_year_1: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True
    )
    completed_gateway_math_year_1: pt.Series[pd.CategoricalDtype] = (
        CompletedGatewayField()
    )
    completed_gateway_english_year_1: pt.Series[pd.CategoricalDtype] = (
        CompletedGatewayField()
    )
    gateway_math_grade_y_1: pt.Series["string"] = GradeField()
    gateway_english_grade_y_1: pt.Series["string"] = GradeField()
    # NOTE: categories set in a parser, which forces "UK" values to null
    attempted_dev_math_y_1: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which forces "UK" values to null
    attempted_dev_english_y_1: pt.Series[pd.CategoricalDtype] = pda.Field(nullable=True)
    completed_dev_math_y_1: pt.Series[pd.CategoricalDtype] = CompletedDevField()
    completed_dev_english_y_1: pt.Series[pd.CategoricalDtype] = CompletedDevField()
    retention: pt.Series["boolean"]
    persistence: pt.Series["boolean"]
    years_to_bachelors_at_cohort_inst: pt.Series["Int8"] = YearsToOfField(nullable=True)
    years_to_associates_or_certificate_at_cohort_inst: pt.Series["Int8"] = (
        YearsToOfField(nullable=True)
    )
    years_to_bachelor_at_other_inst: pt.Series["Int8"] = YearsToOfField(nullable=True)
    years_to_associates_or_certificate_at_other_inst: pt.Series["Int8"] = (
        YearsToOfField(nullable=True)
    )
    years_of_last_enrollment_at_cohort_institution: pt.Series["int32"] = (
        YearsToOfField()
    )
    years_of_last_enrollment_at_other_institution: pt.Series["int32"] = YearsToOfField()
    time_to_credential: pt.Series["Float32"] = pda.Field(nullable=True)
    # NOTE: categories set in a parser, which assigns "UK" values to null
    reading_placement: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True
    )
    special_program: t.Optional[pt.Series["string"]] = pda.Field(nullable=True)
    naspa_first_generation: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True,
        dtype_kwargs={"categories": ["-1", "0", "1", "2", "3", "4", "5", "6"]},
    )
    # NOTE: categories set in a parser, which assigns "UK" values to null
    incarcerated_status: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True
    )
    # NOTE: categories set in a parser, which forces "-1" / "-1.0" values to null
    military_status: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True
    )
    employment_status: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True, dtype_kwargs={"categories": ["-1", "0", "1", "2", "3", "4"]}
    )
    # NOTE: categories set in a parser, which assigns "UK" values to null
    disability_status: t.Optional[pt.Series[pd.CategoricalDtype]] = pda.Field(
        nullable=True
    )
    foreign_language_completion: t.Optional[pt.Series["string"]] = pda.Field(
        nullable=True
    )
    first_year_to_bachelors_at_cohort_inst: pt.Series["Int8"] = YearsToOfField(
        nullable=True
    )
    first_year_to_associates_or_certificate_at_cohort_inst: pt.Series["Int8"] = (
        YearsToOfField(nullable=True)
    )
    first_year_to_bachelor_at_other_inst: pt.Series["Int8"] = YearsToOfField(
        nullable=True
    )
    first_year_to_associates_or_certificate_at_other_inst: pt.Series["Int8"] = (
        YearsToOfField(nullable=True)
    )
    program_of_study_year_1: pt.Series["string"] = pda.Field(nullable=True)
    most_recent_bachelors_at_other_institution_state: pt.Series["string"] = pda.Field(
        nullable=True
    )
    most_recent_associates_or_certificate_at_other_institution_state: pt.Series[
        "string"
    ] = pda.Field(nullable=True)
    most_recent_last_enrollment_at_other_institution_state: pt.Series["string"] = (
        pda.Field(nullable=True)
    )
    first_bachelors_at_other_institution_state: pt.Series["string"] = pda.Field(
        nullable=True
    )
    first_associates_or_certificate_at_other_institution_state: pt.Series["string"] = (
        pda.Field(nullable=True)
    )
    most_recent_bachelors_at_other_institution_carnegie: pt.Series["string"] = (
        pda.Field(nullable=True)
    )
    most_recent_associates_or_certificate_at_other_institution_carnegie: pt.Series[
        "string"
    ] = pda.Field(nullable=True)
    most_recent_last_enrollment_at_other_institution_carnegie: pt.Series["string"] = (
        pda.Field(nullable=True)
    )
    first_bachelors_at_other_institution_carnegie: pt.Series["string"] = pda.Field(
        nullable=True
    )
    first_associates_or_certificate_at_other_institution_carnegie: pt.Series[
        "string"
    ] = pda.Field(nullable=True)
    most_recent_bachelors_at_other_institution_locale: pt.Series[pd.StringDtype] = (
        pda.Field(nullable=True)
    )
    most_recent_associates_or_certificate_at_other_institution_locale: pt.Series[
        pd.StringDtype
    ] = pda.Field(nullable=True)
    most_recent_last_enrollment_at_other_institution_locale: pt.Series[
        pd.StringDtype
    ] = pda.Field(nullable=True)
    first_bachelors_at_other_institution_locale: pt.Series[pd.StringDtype] = pda.Field(
        nullable=True
    )
    first_associates_or_certificate_at_other_institution_locale: pt.Series[
        pd.StringDtype
    ] = pda.Field(nullable=True)
    # added in 2025-01
    years_to_latest_associates_at_cohort_inst: t.Optional[pt.Series["Int8"]] = (
        YearsToOfField(nullable=True)
    )
    years_to_latest_certificate_at_cohort_inst: t.Optional[pt.Series["Int8"]] = (
        YearsToOfField(nullable=True)
    )
    years_to_latest_associates_at_other_inst: t.Optional[pt.Series["Int8"]] = (
        YearsToOfField(nullable=True)
    )
    years_to_latest_certificate_at_other_inst: t.Optional[pt.Series["Int8"]] = (
        YearsToOfField(nullable=True)
    )
    first_year_to_associates_at_cohort_inst: t.Optional[pt.Series["Int8"]] = (
        YearsToOfField(nullable=True)
    )
    first_year_to_certificate_at_cohort_inst: t.Optional[pt.Series["Int8"]] = (
        YearsToOfField(nullable=True)
    )
    first_year_to_associates_at_other_inst: t.Optional[pt.Series["Int8"]] = (
        YearsToOfField(nullable=True)
    )
    first_year_to_certificate_at_other_inst: t.Optional[pt.Series["Int8"]] = (
        YearsToOfField(nullable=True)
    )

    @pda.dataframe_parser
    def rename_student_id_col(cls, df):
        if "study_id" in df.columns:
            LOGGER.info("renaming 'study_id' column => 'student_id'")
            df = df.rename(columns={"study_id": "student_id"}).astype(
                {"student_id": "string"}
            )
        elif "student_guid" in df.columns:
            LOGGER.info("renaming 'student_guid' column => 'student_id'")
            df = df.rename(columns={"student_guid": "student_id"}).astype(
                {"student_id": "string"}
            )
        return df

    @pda.parser(
        "student_age",
        "race",
        "ethnicity",
        "gender",
        "first_gen",
        "credential_type_sought_year_1",
        "most_recent_bachelors_at_other_institution_locale",
        "most_recent_associates_or_certificate_at_other_institution_locale",
        "most_recent_last_enrollment_at_other_institution_locale",
        "first_bachelors_at_other_institution_locale",
        "first_associates_or_certificate_at_other_institution_locale",
    )
    def strip_and_uppercase_strings(cls, series):
        return series.str.strip().str.upper()

    @pda.parser("enrollment_type")
    def set_enrollment_type_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(
            ["FIRST-TIME", "RE-ADMIT", "TRANSFER-IN"]
        )

    @pda.parser("enrollment_intensity_first_term")
    def set_enrollment_intensity_first_term_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(
            ["FULL-TIME", "PART-TIME"]
        )

    @pda.parser("math_placement", "english_placement", "reading_placement")
    def set_subj_placement_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(["C", "N"])

    @pda.parser("dual_and_summer_enrollment")
    def set_dual_and_summer_enrollment_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(
            ["DE", "SE", "DS"]
        )

    @pda.parser("pell_status_first_year")
    def set_pell_status_first_year_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(["Y", "N"])

    @pda.parser("gateway_math_status", "gateway_english_status")
    def set_gateway_math_english_status_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(["R", "N"])

    @pda.parser("attempted_gateway_math_year_1", "attempted_gateway_english_year_1")
    def set_attempted_gateway_math_english_year_1_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(["Y", "N"])

    @pda.parser("attempted_dev_math_y_1", "attempted_dev_english_y_1")
    def set_attempted_dev_math_english_y_1_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(["Y", "N", "NA"])

    @pda.parser("incarcerated_status")
    def set_incarcerated_status_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(["Y", "P", "N"])

    @pda.parser("disability_status")
    def set_disability_status_categories(cls, series):
        return _strip_upper_strings_to_cats(series).cat.set_categories(["Y", "N"])

    @pda.parser("military_status")
    def set_military_status_categories(cls, series):
        return (
            series.astype("Float32")
            .astype("Int8")
            .astype("category")
            .cat.set_categories(["0", "1", "2"])
        )

    @pda.parser(
        "years_to_associates_or_certificate_at_cohort_inst",
        "years_to_bachelors_at_cohort_inst",
        "years_to_associates_or_certificate_at_other_inst",
        "years_to_bachelor_at_other_inst",
        "first_year_to_associates_or_certificate_at_cohort_inst",
        "first_year_to_bachelors_at_cohort_inst",
        "first_year_to_associates_or_certificate_at_other_inst",
        "first_year_to_bachelor_at_other_inst",
        "years_to_latest_associates_at_cohort_inst",
        "years_to_latest_certificate_at_cohort_inst",
        "years_to_latest_associates_at_other_inst",
        "years_to_latest_certificate_at_other_inst",
        "first_year_to_associates_at_cohort_inst",
        "first_year_to_certificate_at_cohort_inst",
        "first_year_to_associates_at_other_inst",
        "first_year_to_certificate_at_other_inst",
    )
    def set_zero_year_values_to_null(cls, series):
        return series.mask(series.eq(0), pd.NA).astype("Int8")

    @pda.check("institution_id", name="check_num_institutions")
    def check_num_institutions(cls, series) -> bool:
        return series.nunique() == 1

    class Config:
        coerce = True
        # "strict" parsing is disabled so we can rename raw identifier cols to student_id
        strict = False
        unique_column_names = True
        add_missing_columns = False
        drop_invalid_rows = False
        unique = ["student_id"]


def _strip_upper_strings_to_cats(series: pd.Series) -> pd.Series:
    return series.str.strip().str.upper().astype("category")



import typing as t
from collections.abc import Collection, Iterable

import pandas as pd

TermType = t.Literal["FALL", "WINTER", "SPRING", "SUMMER"]
IntensityTimeLimitsType = dict[str, tuple[float, t.Literal["year", "term"]]]
CheckpointType = t.Union[pd.DataFrame, t.Callable[[pd.DataFrame], pd.DataFrame]]
CheckpointTypeType = t.Literal[
    "nth",
    "first",
    "last",
    "first_at_num_credits_earned",
    "first_within_cohort",
    "last_in_enrollment_year",
]
TargetTypeType = t.Literal["graduation", "retention", "credits_earned"]


def to_list(value: t.Any) -> list:
    """Cast ``value`` into a list, regardless of its type."""
    if isinstance(value, list):
        return value
    elif isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
        return list(value)
    else:
        return [value]


def is_collection_but_not_string(value: t.Any) -> bool:
    """Check if ``value`` is a collection, excluding strings (and bytes)."""
    return isinstance(value, Collection) and not isinstance(value, (str, bytes))



import logging
from collections.abc import Collection

import numpy as np
import pandas as pd

from ... import utils

LOGGER = logging.getLogger(__name__)


def select_students_by_attributes(
    df: pd.DataFrame,
    *,
    student_id_cols: str | list[str] = "student_id",
    **criteria: object | Collection[object],
) -> pd.DataFrame:
    """
    Select distinct students in ``df`` with specified attributes,
    according to the logical "AND" of one or more selection criteria.

    Args:
        df: Student-term dataset.
        student_id_cols: Column(s) that uniquely identify students in ``df`` ,
            used to drop duplicates.
        **criteria: Column name in ``df`` mapped to one or multiple values
            that it must equal in order for the corresponding student to be selected.
            Multiple criteria are combined with a logical "AND". For example,
            ``enrollment_type="FIRST-TIME", enrollment_intensity_first_term=["FULL-TIME", "PART-TIME"]``
            selects first-time students who enroll at full- or part-time intensity.

    Warning:
        This assumes that students' attribute values are the same across all terms.
        If that assumption is violated, this code actually checks if the student
        satisfied all of the criteria during _any_ of their terms.
    """
    if not criteria:
        raise ValueError("one or more selection criteria must be specified")

    student_id_cols = utils.types.to_list(student_id_cols)
    nuq_students_in = df.groupby(by=student_id_cols, sort=False).ngroups
    is_selecteds = [
        df[key].isin(set(val))  # type: ignore
        if utils.types.is_collection_but_not_string(val)
        else df[key].eq(val).fillna(value=False)  # type: ignore
        for key, val in criteria.items()
    ]
    for (key, val), is_selected in zip(criteria.items(), is_selecteds):
        nuq_students_criterion = (
            df.loc[is_selected, student_id_cols]
            .groupby(by=student_id_cols, sort=False)
            .ngroups
        )
        _log_selection(
            nuq_students_in, nuq_students_criterion, f"{key}={val} criterion ..."
        )
    is_selected = np.logical_and.reduce(is_selecteds)
    df_selected = (
        df.loc[is_selected, student_id_cols + list(criteria.keys())]
        # df is at student-term level; get student ids from first selected terms only
        .drop_duplicates(subset=student_id_cols, ignore_index=True)
        .set_index(student_id_cols)
    )
    _log_selection(nuq_students_in, len(df_selected), "all criteria")
    return df_selected


def _log_selection(
    nunique_students_in: int, nunique_students_out: int, case: str
) -> None:
    LOGGER.info(
        "%s out of %s (%s%%) students selected by %s",
        nunique_students_out,
        nunique_students_in,
        round(100 * nunique_students_out / nunique_students_in, 1),
        case,
    )



import functools as ft
import typing as t

import numpy as np
import pandas as pd

from .... import utils
from ... import checkpoints
from . import shared


def compute_target(
    df: pd.DataFrame,
    *,
    intensity_time_limits: utils.types.IntensityTimeLimitsType,
    years_to_degree_col: str,
    num_terms_in_year: int = 4,
    max_term_rank: int | t.Literal["infer"] = "infer",
    student_id_cols: str | list[str] = "student_id",
    enrollment_intensity_col: str = "student_term_enrollment_intensity",
    enrollment_year_col: str = "year_of_enrollment_at_cohort_inst",
    term_is_pre_cohort_col: str = "term_is_pre_cohort",
    term_rank_col: str = "term_rank",
) -> pd.Series:
    """
    Compute *non* graduation target for each distinct student in ``df`` , for which
    intensity-specific time limits determine if graduation is "on-time"; null values
    are assigned when target can't be computed based on the time spanned in the dataset.

    Args:
        df: Student-term dataset.
        intensity_time_limits: Mapping of enrollment intensity value (e.g. "FULL-TIME")
            to the maximum number of years or terms considered to be an "on-time" graduation
            (e.g. [4.0, "year"], [12.0, "term"]), measured from the first term of enrollment
            within students' cohorts. To use the same time limit for all students,
            regardless of enrollment intensity, specify "*" as the only key.
        years_to_degree_col: Column whose values give the year of enrollment at cohort inst
            in which students *first* earned a particular degree; for example,
            "first_year_to_bachelors_at_cohort_inst" for Bachelors-seeking students.
            Note: If ``df`` has one row per student-term, it's assumed that years-to-degree
            values are the same across all rows, and we simply use the first.
        num_terms_in_year: Number of academic terms in one academic year,
            used to convert from term-based time limits to year-based time limits;
            default value assumes FALL, WINTER, SPRING, and SUMMER terms.
        max_term_rank: Maximum term rank value in the full dataset ``df`` , either inferred
            from ``df[term_rank_col]`` itself or as a manually specified value which
            may be different from the actual max value in ``df`` , depending on use case.
        student_id_cols: One or multiple columns uniquely identifying students.
        enrollment_intensity_col: Column whose values give students' "enrollment intensity"
            (usually either "FULL-TIME" or "PART-TIME"), for which the most common
            value per student is used when comparing against intensity-specific time limits.
        enrollment_year_col: Column whose values give students' "current" year of enrollment
            at the cohort inst as of the given row, used to filter rows to pre-graduation
            when determining students' most common enrollment intensity.
        term_is_pre_cohort_col
        term_rank_col: Column whose values give the absolute integer ranking of a given
            term within the full dataset ``df`` .

    See Also:
        - :func:`shared.get_students_with_max_target_term_in_dataset()`
        - :func:`checkpoints.pdp.first_student_terms_within_cohort()`
    """
    student_id_cols = utils.types.to_list(student_id_cols)
    # we want a target for every student in input df; this will ensure it
    df_distinct_students = df[student_id_cols].drop_duplicates(ignore_index=True)
    # get most common intensity value per student across all pre-graduation terms
    # and *first* years to degree value (we assume this is the same across all terms)
    # use this as reference data for computing target variable
    df_student_intensities = (
        df.loc[
            df[enrollment_year_col]
            .le(df[years_to_degree_col].astype("Int8"))
            .fillna(True),
            :,
        ]
        .groupby(by=student_id_cols, as_index=False)
        .agg(enrollment_intensity=(enrollment_intensity_col, _mode_aggfunc))
    )
    df_student_years_to_degree = df.groupby(by=student_id_cols, as_index=False).agg(
        years_to_degree=(years_to_degree_col, "first")
    )
    df_ref = pd.merge(
        df_student_intensities,
        df_student_years_to_degree,
        how="inner",
        on=student_id_cols,
    )
    # convert from term limits to year limits, as needed
    intensity_num_years = utils.misc.convert_intensity_time_limits(
        "year", intensity_time_limits, num_terms_in_year=num_terms_in_year
    )
    # compute all intensity/year boolean arrays separately
    # then combine with a logical OR
    targets = [
        (
            # enrollment intensity is equal to specified value or "*" given as intensity
            (df_ref["enrollment_intensity"].eq(intensity) | (intensity == "*"))
            & (
                # student graduated after max num years allowed
                (df_ref["years_to_degree"]).gt(num_years)
                # or never graduated at all
                | df_ref["years_to_degree"].isna()
            )
        )
        for intensity, num_years in intensity_num_years.items()
    ]
    target = np.logical_or.reduce(targets)
    # assign True to all students passing intensity/term condition(s) above
    df_target_true = (
        df_ref.loc[target, student_id_cols]
        .assign(target=True)
        .astype({"target": "boolean"})
    )
    # get all students for which a target label can accurately be computed
    # i.e. the data in df covers their last "on-time" graduation term
    df_labelable_students = shared.get_students_with_max_target_term_in_dataset(
        df,
        checkpoint=ft.partial(
            checkpoints.pdp.first_student_terms_within_cohort,
            term_is_pre_cohort_col=term_is_pre_cohort_col,
            student_id_cols=student_id_cols,
            sort_cols=term_rank_col,
            include_cols=(student_id_cols + [term_rank_col, enrollment_intensity_col]),
        ),
        intensity_time_limits=intensity_time_limits,
        max_term_rank=max_term_rank,
        num_terms_in_year=num_terms_in_year,
        student_id_cols=student_id_cols,
        enrollment_intensity_col=enrollment_intensity_col,
        term_rank_col=term_rank_col,
    )
    df_labeled = (
        # match positive labels to label-able students
        pd.merge(df_labelable_students, df_target_true, on=student_id_cols, how="left")
        # assign False to all label-able students not already assigned True
        .fillna({"target": False})
        # structure so student-ids as index, target as only column
        .set_index(student_id_cols)
    )
    df_all_student_targets = (
        # assign null target to all students
        df_distinct_students.assign(target=pd.Series(pd.NA, dtype="boolean"))
        # structure so student-ids as index, target as only column
        .set_index(student_id_cols)
    )
    # update null targets in-place with bool targets on matching student-id indexes
    df_all_student_targets.update(df_labeled)
    # #drop if target is uncalculable (null)
    df_all_student_targets["target"] = (
        df_all_student_targets["target"].astype("boolean").dropna()
    )
    # return as a series with target as values and student ids as index
    return df_all_student_targets.loc[:, "target"].dropna()


def _mode_aggfunc(ser: pd.Series, *, dropna: bool = True) -> object:
    mode = ser.mode(dropna=dropna)
    return mode.iat[0] if not mode.empty else pd.NA



import json
import pandas as pd
import pytest
import uuid
import os

from student_success_tool.ingestion_validation.validation import (
    normalize_col,
    load_json,
    merge_model_columns,
    build_schema,
    validate_dataset,
    HardValidationError,
)
from student_success_tool.ingestion_validation.logger import SimpleLogger


@pytest.fixture
def base_schema_file(tmp_path):
    schema = {
        "base": {
            "data_models": {
                "student": {
                    "columns": {
                        "student_id": {
                            "dtype": "string",
                            "coerce": True,
                            "nullable": False,
                            "required": True,
                            "aliases": ["guid", "study_id"],
                            "checks": [
                                {
                                    "type": "str_length",
                                    "args": [],
                                    "kwargs": {"min_value": 3},
                                }
                            ],
                        },
                        "age": {
                            "dtype": "string",
                            "coerce": True,
                            "nullable": False,
                            "required": False,
                            "aliases": [],
                            "checks": [
                                {
                                    "type": "str_length",
                                    "args": [],
                                    "kwargs": {"min_value": 2},
                                }
                            ],
                        },
                        "disability_status": {
                            "dtype": "string",
                            "coerce": True,
                            "nullable": True,
                            "required": False,
                            "aliases": [],
                            "checks": [],
                        },
                    }
                }
            }
        }
    }
    p = tmp_path / "base_schema.json"
    p.write_text(json.dumps(schema))
    return str(p)


@pytest.fixture
def ext_schema_file(tmp_path):
    ext = {
        "institutions": {
            "institution": {
                "data_models": {
                    "student": {
                        "columns": {
                            "enrollment_type": {
                                "dtype": "string",
                                "coerce": True,
                                "nullable": True,
                                "required": False,
                                "aliases": ["enroll_type"],
                                "checks": [],
                            }
                        }
                    }
                }
            }
        }
    }
    p = tmp_path / "ext_schema.json"
    p.write_text(json.dumps(ext))
    return str(p)


@pytest.fixture
def test_logger(tmp_path) -> SimpleLogger:
    log_file = tmp_path / f"{uuid.uuid4().hex}_validation.log"
    logger = SimpleLogger(log_path=str(log_file), institution_id="test")
    yield logger
    logger.close()


def test_normalize_col():
    assert normalize_col("  student id ") == "student_id"
    assert normalize_col("ENROLLMENT date") == "enrollment_date"


def test_load_json_success(tmp_path):
    data = {"a": 1}
    p = tmp_path / "file.json"
    p.write_text(json.dumps(data))
    assert load_json(str(p)) == data


def test_load_json_not_found(tmp_path):
    with pytest.raises(FileNotFoundError):
        load_json(str(tmp_path / "missing.json"))


def test_load_json_malformed(tmp_path):
    bad = tmp_path / "bad.json"
    bad.write_text("not json")
    with pytest.raises(FileNotFoundError):
        load_json(str(bad))


def test_merge_model_columns_base_only(base_schema_file):
    base = json.load(open(base_schema_file))
    merged = merge_model_columns(
        base, extension_schema=None, institution="institution", model="student"
    )
    assert set(merged) == {"student_id", "age", "disability_status"}


def test_merge_model_columns_with_extension(base_schema_file, ext_schema_file):
    base = json.load(open(base_schema_file))
    ext = json.load(open(ext_schema_file))
    merged = merge_model_columns(base, ext, "institution", "student")
    assert set(merged) == {"student_id", "age", "disability_status", "enrollment_type"}


def test_merge_model_columns_missing_model(base_schema_file):
    base = json.load(open(base_schema_file))
    with pytest.raises(KeyError):
        merge_model_columns(base, None, "institution", "unknown_model")


def test_build_schema_and_validate(base_schema_file):
    df = pd.DataFrame(
        {
            "student_id": ["ABC123", "DEF456"],
            "age": ["21+", "<21"],
            "disability_status": ["Y", None],
        }
    )
    base = json.load(open(base_schema_file))
    specs = base["base"]["data_models"]["student"]["columns"]
    schema = build_schema(specs)
    validated = schema.validate(df)
    pd.testing.assert_frame_equal(validated, df, check_dtype=False)


def test_validate_dataset_extra_columns(
    monkeypatch, tmp_path, base_schema_file, test_logger
):
    df = pd.DataFrame(
        {
            "student_id": ["ABC123"],
            "age": ["21+"],
            "disability_status": ["Y"],
            "unexpected": [1],
        }
    )
    monkeypatch.setattr(
        "student_success_tool.ingestion_validation.validation.load_json",
        lambda path: json.load(open(base_schema_file)),
    )
    with pytest.raises(HardValidationError) as exc:
        validate_dataset(
            df,
            models="student",
            institution_id="pdp",
            logger=test_logger,
        )
    err = exc.value
    assert err.extra_columns == ["unexpected"]
    assert err.missing_required == []


def test_validate_dataset_missing_required(monkeypatch, base_schema_file, test_logger):
    df = pd.DataFrame({"age": ["21+"], "disability_status": ["Y"]})
    monkeypatch.setattr(
        "student_success_tool.ingestion_validation.validation.load_json",
        lambda path: json.load(open(base_schema_file)),
    )
    with pytest.raises(HardValidationError) as exc:
        validate_dataset(
            df,
            models="student",
            institution_id="institution",
            logger=test_logger,
        )
    err = exc.value
    assert err.missing_required == ["student_id"]


def test_validate_dataset_schema_errors(monkeypatch, base_schema_file, test_logger):
    df = pd.DataFrame(
        {
            "student_id": ["AB"],
            "age": ["21+"],
            "disability_status": ["Y"],
        }
    )
    monkeypatch.setattr(
        "student_success_tool.ingestion_validation.validation.load_json",
        lambda path: json.load(open(base_schema_file)),
    )
    with pytest.raises(HardValidationError) as exc:
        validate_dataset(
            df,
            models="student",
            institution_id="institution",
            logger=test_logger,
        )
    err = exc.value
    assert err.schema_errors is not None
    assert isinstance(err.schema_errors, list)


def test_validate_dataset_soft_pass(
    monkeypatch, base_schema_file, ext_schema_file, test_logger
):
    df = pd.DataFrame({"student_id": ["ABC123"], "disability_status": ["N"]})

    # Always return base or ext schema by inspecting the path basename
    def mocked_load_json(path):
        if (
            path
            == "/Volumes/staging_sst_01/institution_bronze/bronze_volume/schema/institution_schema_extension.json"
        ):
            return json.load(open(ext_schema_file))
        return json.load(open(base_schema_file))

    # Patch both `load_json` and `os.path.exists` to simulate presence of extension schema
    monkeypatch.setattr(
        "student_success_tool.ingestion_validation.validation.load_json",
        mocked_load_json,
    )

    monkeypatch.setattr(
        "student_success_tool.ingestion_validation.validation.os.path.exists",
        lambda path: True if "schema_extension.json" in path else os.path.exists(path),
    )

    result = validate_dataset(
        df,
        models="student",
        institution_id="institution",
        logger=test_logger,
    )

    assert result["validation_status"] == "passed_with_soft_errors"
    assert set(result["missing_optional"]) == {"age", "enrollment_type"}


def test_validate_dataset_csv_input(
    tmp_path, monkeypatch, base_schema_file, test_logger
):
    df = pd.DataFrame(
        {"student_id": ["ABC123"], "age": ["21+"], "disability_status": ["N"]}
    )
    csv = tmp_path / "in.csv"
    df.to_csv(csv, index=False)

    monkeypatch.setattr(
        "student_success_tool.ingestion_validation.validation.load_json",
        lambda path: json.load(open(base_schema_file)),
    )
    result = validate_dataset(
        str(csv),
        models="student",
        institution_id="institution",
        logger=test_logger,
    )
    assert result["validation_status"] in ("passed", "passed_with_soft_errors")



import functools as ft
import logging
import typing as t

import pandas as pd
import pyspark.sql

from ... import utils
from .. import read

try:
    import pandera as pda
except ModuleNotFoundError:
    from ... import utils

    utils.databricks.mock_pandera()

    import pandera as pda

LOGGER = logging.getLogger(__name__)


def read_raw_course_data(
    *,
    table_path: t.Optional[str] = None,
    file_path: t.Optional[str] = None,
    schema: t.Optional[type[pda.DataFrameModel]] = None,
    dttm_format: str = "%Y%m%d",
    converter_func: t.Optional[t.Callable[[pd.DataFrame], pd.DataFrame]] = None,
    spark_session: t.Optional[pyspark.sql.SparkSession] = None,
    **kwargs: object,
) -> pd.DataFrame:
    """
    Read raw PDP course data from table (in Unity Catalog) or file (in CSV format),
    and parse+validate it via ``schema`` .

    Args:
        table_path
        file_path
        schema: "DataFrameModel", such as those specified in :mod:`schemas` ,
            used to parse and validate the raw data. If None, parsing/validation
            is skipped, and the raw data is returned as-is.
        dttm_format: Datetime format for "Course Begin/End Date" columns.
        converter_func: If the raw data is incompatible with ``schema`` ,
            provide a function that takes the raw dataframe as its sole input,
            performs whatever (minimal) transformations necessary to bring the data
            into line with ``schema`` , and then returns it. This converted dataset
            will then be passed into ``schema`` , if specified.
            NOTE: Allowances for minor differences in the data should be implemented
            on the school-specific schema class directly. This function is intended
            to handle bigger problems, such as duplicate ids or borked columns.
        spark_session: Required if reading data from ``table_path`` , and optional
            if reading data from ``file_path`` .
        **kwargs: Additional arguments passed as-is into underlying read func.
            Note that raw data is always read in as "string" dtype, then coerced
            into the correct dtypes using ``schema`` .

    See Also:
        - :func:`read_data_from_csv_file()`
        - :func:`read_data_from_delta_table()`

    References:
        - https://help.studentclearinghouse.org/pdp/knowledge-base/course-level-analysis-ready-file-data-dictionary
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        - https://pandera.readthedocs.io/en/stable/reference/generated/pandera.api.dataframe.model.DataFrameModel.html#pandera.api.dataframe.model.DataFrameModel.validate
    """
    if not bool(table_path) ^ bool(file_path):
        raise ValueError("exactly one of table_path or file_path must be specified")
    elif table_path is not None and spark_session is None:
        raise ValueError("spark session must be given when reading data from table")

    df = (
        read.from_csv_file(file_path, spark_session, **kwargs)  # type: ignore
        if file_path
        else read.from_delta_table(table_path, spark_session)  # type: ignore
    )
    # apply to the data what pandera calls "parsers" before validation
    # ideally, all these operations would be dataframe parsers on the schema itself
    # but pandera applies core before custom parsers under the hood :/
    df = (
        # standardize column names
        df.rename(columns=utils.misc.convert_to_snake_case)
        # standardize certain column values
        .assign(
            # uppercase string values for some cols to avoid case inconsistency later on
            **{
                col: ft.partial(_uppercase_string_values, col=col)
                for col in ("academic_term",)
            }
            # help pandas to parse non-standard datetimes... read_csv() struggles
            | {
                col: ft.partial(_parse_dttm_values, col=col, fmt=dttm_format)
                for col in ("course_begin_date", "course_end_date")
            }
        )
    )
    return _maybe_convert_maybe_validate_data(df, converter_func, schema)


def read_raw_cohort_data(
    *,
    table_path: t.Optional[str] = None,
    file_path: t.Optional[str] = None,
    schema: t.Optional[type[pda.DataFrameModel]] = None,
    converter_func: t.Optional[t.Callable[[pd.DataFrame], pd.DataFrame]] = None,
    spark_session: t.Optional[pyspark.sql.SparkSession] = None,
    **kwargs: object,
) -> pd.DataFrame:
    """
    Read raw PDP cohort data from table (in Unity Catalog) or file (in CSV format),
    and parse+validate it via ``schema`` .

    Args:
        table_path
        file_path
        schema: "DataFrameModel", such as those specified in :mod:`schemas` ,
            used to parse and validate the raw data. If None, parsing/validation
            is skipped, and the raw data is returned as-is.
        converter_func: If the raw data is incompatible with ``schema`` ,
            provide a function that takes the raw dataframe as its sole input,
            performs whatever (minimal) transformations necessary to bring the data
            into line with ``schema`` , and then returns it. This converted dataset
            will then be passed into ``schema`` , if specified.
            NOTE: Allowances for minor differences in the data should be implemented
            on the school-specific schema class directly. This function is intended
            to handle bigger problems, such as duplicate ids or borked columns.
        spark_session: Required if reading data from ``table_path`` , and optional
            if reading data from ``file_path`` .
        **kwargs: Additional arguments passed as-is into underlying read func.
            Note that raw data is always read in as "string" dtype, then coerced
            into the correct dtypes using ``schema`` .

    See Also:
        - :func:`read_data_from_csv_file()`
        - :func:`read_data_from_delta_table()`

    References:
        - https://help.studentclearinghouse.org/pdp/knowledge-base/cohort-level-analysis-ready-file-data-dictionary
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        - https://pandera.readthedocs.io/en/stable/reference/generated/pandera.api.dataframe.model.DataFrameModel.html#pandera.api.dataframe.model.DataFrameModel.validate
    """
    if not bool(table_path) ^ bool(file_path):
        raise ValueError("exactly one of table_path or file_path must be specified")
    elif table_path is not None and spark_session is None:
        raise ValueError("spark session must be given when reading data from table")

    df = (
        read.from_csv_file(file_path, spark_session, **kwargs)  # type: ignore
        if file_path
        else read.from_delta_table(table_path, spark_session)  # type: ignore
    )
    # apply to the data what pandera calls "parsers" before validation
    # ideally, all these operations would be dataframe parsers on the schema itself
    # but pandera applies core before custom parsers under the hood :/
    df = (
        # standardize column names
        df.rename(columns=utils.misc.convert_to_snake_case)
        # standardize column values
        .assign(
            # uppercase string values for some cols to avoid case inconsistency later on
            # for practical reasons, this is the only place where it's easy to do so
            **{
                col: ft.partial(_uppercase_string_values, col=col)
                for col in ("cohort_term",)
            }
            # replace "UK" with null in GPA cols, so we can coerce to float via schema
            | {
                col: ft.partial(_replace_values_with_null, col=col, to_replace="UK")
                for col in ("gpa_group_term_1", "gpa_group_year_1")
            }
            # help pandas to coerce string "1"/"0" values into True/False
            | {
                col: ft.partial(_cast_to_bool_via_int, col=col)
                for col in ("retention", "persistence")
            }
        )
    )
    return _maybe_convert_maybe_validate_data(df, converter_func, schema)


def _maybe_convert_maybe_validate_data(
    df: pd.DataFrame,
    converter_func: t.Optional[t.Callable[[pd.DataFrame], pd.DataFrame]] = None,
    schema: t.Optional[type[pda.DataFrameModel]] = None,
) -> pd.DataFrame:
    # HACK: we're hiding this pandera import here so databricks doesn't know about it
    # pandera v0.23+ pulls in pandas v2.1+ while databricks runtimes are stuck in v1.5
    # resulting in super dumb dependency errors when loading automl trained models
    import pandera.errors

    if converter_func is not None:
        LOGGER.info("applying %s converter to raw data", converter_func)
        df = converter_func(df)
    if schema is None:
        return df
    else:
        try:
            df_validated = schema.validate(df, lazy=True)
            assert isinstance(df_validated, pd.DataFrame)
            return df_validated
        except pandera.errors.SchemaErrors:
            LOGGER.error("unable to parse/validate raw data")
            raise


def _parse_dttm_values(df: pd.DataFrame, *, col: str, fmt: str) -> pd.Series:
    return pd.to_datetime(df[col], format=fmt)


def _uppercase_string_values(df: pd.DataFrame, *, col: str) -> pd.Series:
    return df[col].str.upper()


def _replace_values_with_null(
    df: pd.DataFrame, *, col: str, to_replace: str | list[str]
) -> pd.Series:
    return df[col].replace(to_replace=to_replace, value=None)


def _cast_to_bool_via_int(df: pd.DataFrame, *, col: str) -> pd.Series:
    return (
        df[col]
        .astype("string")
        .map(
            {
                "1": True,
                "0": False,
                "True": True,
                "False": False,
                "true": True,
                "false": False,
            }
        )
        .astype("boolean")
    )



import logging
import typing as t

import numpy as np
import pandas as pd

from .... import utils

LOGGER = logging.getLogger(__name__)


def get_students_with_max_target_term_in_dataset(
    df: pd.DataFrame,
    *,
    checkpoint: pd.DataFrame | t.Callable[[pd.DataFrame], pd.DataFrame],
    intensity_time_limits: utils.types.IntensityTimeLimitsType,
    max_term_rank: int | t.Literal["infer"] = "infer",
    num_terms_in_year: int = 4,
    student_id_cols: str | list[str] = "student_id",
    enrollment_intensity_col: str = "student_term_enrollment_intensity",
    term_rank_col: str = "term_rank",
) -> pd.DataFrame:
    """
    Get set of students in ``checkpoint`` data for which ``df`` includes
    the last possible term within specified time limits, depending on the student's
    enrollment intensity in checkpoint term, such that a target could be known for sure.

    Args:
        df: Student-term dataset.
        checkpoint: "Checkpoint" from which time limits to target term are measured,
            typically either the first enrolled term or the first term above an intermediate
            number of credits earned; may be given as a data frame with one row per student,
            or as a callable that takes ``df`` as input and returns all checkpoint terms.
        intensity_time_limits: Mapping of enrollment intensity value (e.g. "FULL-TIME")
            to the maximum number of years or terms considered to be "on-time" for
            the target number of credits earned (e.g. [4.0, "year"], [12.0, "term"]),
            where the numeric values are for the time between "checkpoint" and "target"
            terms. Passing special "*" as the only key applies the corresponding time limits
            to all students, regardless of intensity.
        max_term_rank: Maximum term rank value in the full dataset ``df`` , either inferred
            from ``df[term_rank_col]`` itself or as a manually specified value which
            may be different from the actual max value in ``df`` , depending on use case.
        num_terms_in_year: Number of academic terms in one academic year,
            used to convert from year-based time limits to term-based time limits;
            default value assumes FALL, WINTER, SPRING, and SUMMER terms.
        student_id_cols: One or multiple columns uniquely identifying students.
        enrollment_intensity_col: Column whose values give students' "enrollment intensity"
            (usually either "FULL-TIME" or "PART-TIME"), for use in applying a time limit.
        term_rank_col: Column whose values give the absolute integer ranking of a given
            term within the full dataset ``df`` .
    """
    if max_term_rank == "infer":
        max_term_rank = int(df[term_rank_col].max())
    assert isinstance(max_term_rank, int)  # type guard

    student_id_cols = utils.types.to_list(student_id_cols)
    nuq_students_in = df.groupby(by=student_id_cols, sort=False).ngroups
    df_ckpt = (
        checkpoint.copy(deep=True)
        if isinstance(checkpoint, pd.DataFrame)
        else checkpoint(df)
    )
    if df_ckpt.groupby(by=student_id_cols).size().gt(1).any():
        raise ValueError("checkpoint df must include exactly 1 row per student")

    intensity_num_terms = utils.misc.convert_intensity_time_limits(
        "term", intensity_time_limits, num_terms_in_year=num_terms_in_year
    )
    if "*" in intensity_num_terms:
        df_ckpt = df_ckpt.assign(
            student_max_term_rank=lambda df: df[term_rank_col]
            + intensity_num_terms["*"]
        )
    else:
        df_ckpt = df_ckpt.assign(
            student_max_term_rank=lambda df: df[term_rank_col]
            + df[enrollment_intensity_col].map(intensity_num_terms)
        )
    df_out = (
        df_ckpt.loc[
            df_ckpt["student_max_term_rank"].le(max_term_rank),
            student_id_cols + ["student_max_term_rank"],
        ]
        # get student ids from first eligible terms only
        .drop_duplicates(subset=student_id_cols, ignore_index=True)
        .set_index(student_id_cols)
        # include relevant other columns, mostly just for debugging
        .assign(max_term_rank=max_term_rank)
        .astype("Int8")
    )
    _log_labelable_students(nuq_students_in, len(df_out))
    return df_out


def get_students_with_second_year_in_dataset(
    df: pd.DataFrame,
    *,
    max_academic_year: str | t.Literal["infer"] = "infer",
    student_id_cols: str | list[str] = "student_id",
    cohort_id_col: str = "cohort_id",
    term_id_col: str = "term_id",
) -> pd.DataFrame:
    """
    Get set of students in ``df`` for which ``df`` includes any records
    for the academic year after the students' cohort year (i.e. their second year);
    effectively, this excludes all students in the cohort from the most recent course year.

    Args:
        df: Student-term dataset.
        max_academic_year: Maximum academic year in the full dataset ``df`` , either inferred
            from ``df[term_id_col]`` itself or as a manually specified value which
            may be different from the actual max value in ``df`` , depending on use case.
            Note: Value must be a string formatted as "YYYY[-YY]".
        student_id_cols: One or multiple columns uniquely identifying students.
        cohort_id_col: Column used to uniquely identify student cohorts.
        term_id_col: Colum used to uniquely identify academic terms.
    """
    if max_academic_year == "infer":
        max_academic_year = _extract_year_from_id(df[term_id_col]).max()
    else:
        max_academic_year = _extract_year_from_id(pd.Series(max_academic_year)).iat[0]
    assert isinstance(max_academic_year, np.integer)  # type guard

    student_id_cols = utils.types.to_list(student_id_cols)
    nuq_students_in = df.groupby(by=student_id_cols, sort=False).ngroups
    df_out = (
        # distinct student ids set as index
        df.groupby(by=student_id_cols, as_index=True)
        # assume every value for a student's cohort id is the same, so "first" is fine
        .agg(student_cohort_id=(cohort_id_col, "first"))
        .assign(
            student_cohort_year=lambda df: _extract_year_from_id(
                df["student_cohort_id"]
            ),
            max_academic_year=max_academic_year,  # type: ignore
        )
        .loc[
            lambda df: df["student_cohort_year"].lt(max_academic_year),
            ["student_cohort_year", "max_academic_year"],
        ]
        .astype("Int32")
    )
    _log_labelable_students(nuq_students_in, len(df_out))
    return df_out


def _log_labelable_students(
    nunique_students_in: int, nunique_students_out: int
) -> None:
    LOGGER.info(
        "%s out of %s (%s%%) students in dataset are able to have targets computed",
        nunique_students_out,
        nunique_students_in,
        round(100 * nunique_students_out / nunique_students_in, 1),
    )


def _extract_year_from_id(ser: pd.Series) -> pd.Series:
    return ser.astype("string").str.extract(r"^(\d{4})", expand=False).astype("Int32")



"""
This script ingests course and cohort data for the Student Success Tool (SST) pipeline.

It reads data from CSV files stored in a Google Cloud Storage (GCS) bucket,
performs schema validation using the `pdp` library, and writes the validated data
to Delta Lake tables in Databricks Unity Catalog.

The script is designed to run within a Databricks environment as a job, leveraging
Databricks utilities for job task values, and Spark session management.

This is a POC script, it is advised to review and tests before using in production.
"""

import logging
import os
import argparse
import sys

from databricks.connect import DatabricksSession
from databricks.sdk.runtime import dbutils
from google.cloud import storage

import student_success_tool.dataio as dataio
import importlib


# Configure logging
logging.basicConfig(level=logging.INFO)
logging.getLogger("py4j").setLevel(logging.WARNING)  # Ignore Databricks logger


class DataIngestionTask:
    """
    Encapsulates the data ingestion logic for the SST pipeline.
    """

    def __init__(
        self,
        args: argparse.Namespace,
        course_converter_func=None,
        cohort_converter_func=None,
    ):
        """
        Initializes the DataIngestionTask with parsed arguments.
        Args:
            args (argparse.Namespace): Parsed command-line arguments.
        """
        self.spark_session = self.get_spark_session()
        self.args = args
        self.storage_client = storage.Client()
        self.bucket = self.storage_client.bucket(self.args.gcp_bucket_name)
        self.course_converter_func = course_converter_func
        self.cohort_converter_func = cohort_converter_func

    def get_spark_session(self) -> DatabricksSession:
        """
        Attempts to create a Spark session.
        Returns:
            DatabricksSession Spark session if successful, raise exception otherwise.
        """
        try:
            spark_session = DatabricksSession.builder.getOrCreate()
            logging.info("Spark session created successfully.")
            return spark_session
        except Exception:
            logging.error(
                "Unable to create Spark session; are you in a Databricks runtime?"
            )
            raise

    def download_data_from_gcs(self, internal_pipeline_path: str) -> tuple[str, str]:
        """
        Downloads course and cohort data from GCS to the internal pipeline directory.

        Args:
            internal_pipeline_path (str): The path to the internal pipeline directory.

        Returns:
            tuple[str, str]: The file paths of the downloaded course and cohort data.
        """
        sst_container_folder = "validated"
        try:
            # Download course data from GCS
            course_blob_name = f"{sst_container_folder}/{self.args.course_file_name}"
            course_blob = self.bucket.blob(course_blob_name)
            course_file_path = f"{internal_pipeline_path}{self.args.course_file_name}"
            course_blob.download_to_filename(course_file_path)
            logging.info("Course data downloaded from GCS: %s", course_file_path)

            # Download cohort data from GCS
            cohort_blob_name = f"{sst_container_folder}/{self.args.cohort_file_name}"
            cohort_blob = self.bucket.blob(cohort_blob_name)
            cohort_file_path = f"{internal_pipeline_path}{self.args.cohort_file_name}"
            cohort_blob.download_to_filename(cohort_file_path)
            logging.info("Cohort data downloaded from GCS: %s", cohort_file_path)

            return course_file_path, cohort_file_path
        except Exception as e:
            logging.error(f"GCS download error: {e}")
            raise

    def read_and_validate_data(self, fpath_course: str, fpath_cohort: str):
        # -> tuple[schemas.RawPDPCourseDataSchema, schemas.RawPDPCohortDataSchema]:
        """
        Reads course and cohort data from CSV files and validates their schemas.

        Args:
            fpath_course (str): Path to the course data file.
            fpath_cohort (str): Path to the cohort data file.

        Returns:
            tuple[schemas.RawPDPCourseDataSchema, schemas.RawPDPCohortDataSchema]:
                Validated course and cohort data.
        """

        # Read data from CSV files into Pandas DataFrames and validate schema
        try:
            df_course = dataio.pdp.raw_data.read_raw_course_data(
                file_path=fpath_course,
                schema=schemas.RawPDPCourseDataSchema,
                dttm_format="ISO8601",
                converter_func=self.course_converter_func,
            )
        except ValueError:
            df_course = dataio.pdp.raw_data.read_raw_course_data(
                file_path=fpath_course,
                schema=schemas.RawPDPCourseDataSchema,
                dttm_format="%Y%m%d.0",
                converter_func=self.course_converter_func,
            )
        except Exception as e:
            logging.error("Error reading the files: %s", e)
            raise

        logging.info("Course data read and schema validated.")
        df_cohort = dataio.pdp.raw_data.read_raw_cohort_data(
            file_path=fpath_cohort,
            schema=schemas.RawPDPCohortDataSchema,
            converter_func=self.cohort_converter_func,
        )
        logging.info("Cohort data read and schema validated.")
        return df_course, df_cohort

    def write_data_to_delta_lake(
        self,
        df_course,
        df_cohort,
    ):
        """
        Writes the validated DataFrames to Delta Lake tables in Unity Catalog.

        Args:
            df_course (schemas.RawPDPCourseDataSchema): Validated course data.
            df_cohort (schemas.RawPDPCohortDataSchema): Validated cohort data.
        """

        catalog = self.args.DB_workspace
        write_schema = f"{self.args.databricks_institution_name}_bronze"
        course_dataset_validated_path = (
            f"{catalog}.{write_schema}.{self.args.db_run_id}_course_dataset_validated"
        )
        cohort_dataset_validated_path = (
            f"{catalog}.{write_schema}.{self.args.db_run_id}_cohort_dataset_validated"
        )
        try:
            dataio.write.to_delta_table(
                df_course,
                course_dataset_validated_path,
                spark_session=self.spark_session,
            )
            logging.info(
                "Course data written to Delta Lake table: %s.%s.%s_course_dataset_validated",
                catalog,
                write_schema,
                self.args.db_run_id,
            )

            dataio.write.to_delta_table(
                df_cohort,
                cohort_dataset_validated_path,
                spark_session=self.spark_session,
            )
            logging.info(
                "Cohort data written to Delta Lake table: %s.%s.%s_cohort_dataset_validated",
                catalog,
                write_schema,
                self.args.db_run_id,
            )

            return course_dataset_validated_path, cohort_dataset_validated_path
        except Exception as e:
            logging.error("Error writing to Delta Lake: %s", e)
            raise

    def verify_delta_lake_write(
        self, course_dataset_validated_path, cohort_dataset_validated_path
    ):
        """
        Verifies the Delta Lake write by reading the data back from the tables.
        """
        try:
            df_course_from_catalog = schemas.RawPDPCourseDataSchema(
                dataio.from_delta_table(
                    course_dataset_validated_path,
                    spark_session=self.spark_session,
                )
            )
            logging.info(
                "Course DataFrame shape from catalog: %s", df_course_from_catalog.shape
            )

            df_cohort_from_catalog = schemas.RawPDPCohortDataSchema(
                dataio.from_delta_table(
                    cohort_dataset_validated_path,
                    spark_session=self.spark_session,
                )
            )
            logging.info(
                "Cohort DataFrame shape from catalog: %s", df_cohort_from_catalog.shape
            )
        except Exception as e:
            logging.error("Error writing to Delta Lake: %s", e)
            raise

    def run(self):
        """
        Executes the data ingestion task.
        """
        raw_files_path = f"{self.args.job_root_dir}/raw_files/"
        # os.makedirs(raw_files_path, exist_ok=True)
        print("raw_files_path:", raw_files_path)
        dbutils.fs.mkdirs(raw_files_path)

        # fpath_course, fpath_cohort = self.download_data_from_gcs(raw_files_path)
        # Hack to get around gcp permissions right now
        fpath_course = f"/Volumes/staging_sst_01/{args.databricks_institution_name}_bronze/bronze_volume/inference_inputs/{self.args.course_file_name}"
        fpath_cohort = f"/Volumes/staging_sst_01/{args.databricks_institution_name}_bronze/bronze_volume/inference_inputs/{self.args.cohort_file_name}"
        df_course, df_cohort = self.read_and_validate_data(fpath_course, fpath_cohort)

        course_dataset_validated_path, cohort_dataset_validated_path = (
            self.write_data_to_delta_lake(df_course, df_cohort)
        )
        self.verify_delta_lake_write(
            course_dataset_validated_path, cohort_dataset_validated_path
        )
        # Setting task variables for downstream tasks
        dbutils.jobs.taskValues.set(
            key="course_dataset_validated_path", value=course_dataset_validated_path
        )
        dbutils.jobs.taskValues.set(
            key="cohort_dataset_validated_path", value=cohort_dataset_validated_path
        )
        dbutils.jobs.taskValues.set(key="job_root_dir", value=self.args.job_root_dir)


def parse_arguments() -> argparse.Namespace:
    """
    Parses command-line arguments.
    Returns:
        argparse.Namespace: An object containing the parsed arguments.
    """
    parser = argparse.ArgumentParser(
        description="Ingest course and cohort data for the SST pipeline."
    )
    parser.add_argument(
        "--DB_workspace", required=True, help="Databricks workspace identifier"
    )
    parser.add_argument(
        "--databricks_institution_name",
        required=True,
        help="Databricksified institution name",
    )
    parser.add_argument(
        "--course_file_name", required=True, help="Name of the course data file"
    )
    parser.add_argument(
        "--cohort_file_name", required=True, help="Name of the cohort data file"
    )
    parser.add_argument(
        "--db_run_id", required=True, help="Databricks job run identifier"
    )
    parser.add_argument(
        "--gcp_bucket_name", required=True, help="Name of the GCP bucket"
    )
    parser.add_argument(
        "--job_root_dir", required=True, help="Folder path to store job output files"
    )
    parser.add_argument(
        "--custom_schemas_path",
        required=False,
        help="Folder path to store custom schemas folders",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_arguments()
    sys.path.append(args.custom_schemas_path)
    sys.path.append(
        f"/Volumes/staging_sst_01/{args.databricks_institution_name}_bronze/bronze_volume/inference_inputs"
    )
    try:
        print("Listdir1", os.listdir("/Workspace/Users"))
        # converter_func = importlib.import_module(f"{args.databricks_institution_name}.dataio")
        converter_func = importlib.import_module("dataio")
        course_converter_func = converter_func.converter_func_course
        cohort_converter_func = converter_func.converter_func_cohort
        logging.info("Running task with custom converter func")
    except ModuleNotFoundError:
        print("Running task without custom converter func")
        course_converter_func = None
        cohort_converter_func = None
        logging.info("Running task without custom converter func")
    try:
        print("sys.path:", sys.path)
        # schemas = importlib.import_module(f"{args.databricks_institution_name}.schemas")
        schemas = importlib.import_module("schemas")
        logging.info("Running task with custom schema")
    except Exception:
        print("Running task with default schema")
        print("Exception", Exception)
        from student_success_tool.dataio.schemas import pdp as schemas

        logging.info("Running task with default schema")

    task = DataIngestionTask(args)
    task.run()



import typing as t

import numpy as np
import pandas as pd

from .... import utils
from ... import checkpoints
from . import shared


def compute_target(
    df: pd.DataFrame,
    *,
    min_num_credits: float,
    checkpoint: pd.DataFrame | t.Callable[[pd.DataFrame], pd.DataFrame],
    intensity_time_limits: utils.types.IntensityTimeLimitsType,
    num_terms_in_year: int = 4,
    max_term_rank: int | t.Literal["infer"] = "infer",
    student_id_cols: str | list[str] = "student_id",
    enrollment_intensity_col: str = "student_term_enrollment_intensity",
    num_credits_col: str = "cumsum_num_credits_earned",
    term_rank_col: str = "term_rank",
) -> pd.Series:
    """
    Compute *insufficient* credits earned target for each distinct student in ``df`` ,
    for which intensity-specific time limits determine if credits earned is "on-time".

    Args:
        df: Student-term dataset.
        min_num_credits: Minimum number of credits earned within specified time limits
            to be considered a *success* => target=False.
        checkpoint: "Checkpoint" from which time limits to target term are determined,
            typically either the first enrolled term or the first term above an intermediate
            number of credits earned; may be given as a data frame with one row per student,
            or as a callable that takes ``df`` as input and returns all checkpoint terms.
        intensity_time_limits: Mapping of enrollment intensity value (e.g. "FULL-TIME")
            to the maximum number of years or terms considered to be "on-time" for
            the target number of credits earned (e.g. [4.0, "year"], [12.0, "term"]),
            where the numeric values are for the time between "checkpoint" and "target"
            terms. Passing special "*" as the only key applies the corresponding time limits
            to all students, regardless of intensity.
        num_terms_in_year: Number of academic terms in one academic year,
            used to convert from year-based time limits to term-based time limits;
            default value assumes FALL, WINTER, SPRING, and SUMMER terms.
        max_term_rank: Maximum term rank value in the full dataset ``df`` , either inferred
            from ``df[term_rank_col]`` itself or as a manually specified value which
            may be different from the actual max value in ``df`` , depending on use case.
        student_id_cols: Columns that uniquely identify students, used for grouping rows.
        enrollment_intensity_col: Column whose values give students' "enrollment intensity"
            (usually either "FULL-TIME" or "PART-TIME"), for which the most common
            value per student is used when comparing against intensity-specific time limits.
        num_credits_col
        term_rank_col: Column whose values give the absolute integer ranking of a given
            term within the full dataset ``df`` .
    """
    student_id_cols = utils.types.to_list(student_id_cols)
    # we want a target for every student in input df; this will ensure it
    df_distinct_students = df[student_id_cols].drop_duplicates(ignore_index=True)
    df_ckpt = (
        checkpoint.copy(deep=True)
        if isinstance(checkpoint, pd.DataFrame)
        else checkpoint(df)
    )
    if df_ckpt.groupby(by=student_id_cols).size().gt(1).any():
        raise ValueError("checkpoint df must include exactly 1 row per student")

    df_tgt = checkpoints.pdp.first_student_terms_at_num_credits_earned(
        df,
        min_num_credits=min_num_credits,
        student_id_cols=student_id_cols,
        sort_cols=term_rank_col,
        num_credits_col=num_credits_col,
        include_cols=[enrollment_intensity_col],
    )
    df_at = pd.merge(
        df_ckpt,
        df_tgt,
        on=student_id_cols,
        how="left",
        suffixes=("_ckpt", "_tgt"),
    )
    # convert from year limits to term limits, as needed
    intensity_num_terms = utils.misc.convert_intensity_time_limits(
        "term", intensity_time_limits, num_terms_in_year=num_terms_in_year
    )
    # compute all intensity/term boolean arrays separately
    # then combine with a logical OR
    tr_col = term_rank_col  # hack, so logic below fits on lines
    targets = [
        (
            # enrollment intensity is equal to a specified value or "*" given as intensity
            (
                df_at[f"{enrollment_intensity_col}_ckpt"].eq(intensity)
                | (intensity == "*")
            )
            & (
                # num terms between target/checkpoint greater than max num allowed
                (df_at[f"{tr_col}_tgt"] - df_at[f"{tr_col}_ckpt"]).gt(num_terms)
                # or they *never* earned enough credits for target
                | df_at[f"{tr_col}_tgt"].isna()
            )
        )
        for intensity, num_terms in intensity_num_terms.items()
    ]
    target = np.logical_or.reduce(targets)
    # assign True to all students passing intensity/year condition(s) above
    df_target_true = (
        df_at.loc[target, student_id_cols]
        .assign(target=True)
        .astype({"target": "boolean"})
    )
    # get all students for which a target label can accurately be computed
    # i.e. the data in df covers their last "on-time" graduation term
    df_labelable_students = shared.get_students_with_max_target_term_in_dataset(
        df,
        checkpoint=df_ckpt,
        intensity_time_limits=intensity_time_limits,
        max_term_rank=max_term_rank,
        num_terms_in_year=num_terms_in_year,
        student_id_cols=student_id_cols,
        enrollment_intensity_col=enrollment_intensity_col,
        term_rank_col=term_rank_col,
    )
    df_labeled = (
        # match positive labels to label-able students
        pd.merge(df_labelable_students, df_target_true, on=student_id_cols, how="left")
        # assign False to all label-able students not already assigned True
        .fillna({"target": False})
        # structure so student-ids as index, target as only column
        .set_index(student_id_cols)
    )
    df_all_student_targets = (
        # assign null target to all students
        df_distinct_students.assign(target=pd.Series(pd.NA, dtype="boolean"))
        # structure so student-ids as index, target as only column
        .set_index(student_id_cols)
    )
    # update null targets in-place with bool targets on matching student-id indexes
    df_all_student_targets.update(df_labeled)
    # #drop if target is uncalculable (null)
    df_all_student_targets["target"] = (
        df_all_student_targets["target"].astype("boolean").dropna()
    )
    # return as a series with target as values and student ids as index
    return df_all_student_targets.loc[:, "target"].dropna()



import logging
import typing as t

import numpy as np
import pandas as pd
import sklearn.utils

LOGGER = logging.getLogger(__name__)

_DEFAULT_SPLIT_LABEL_FRACS = {"train": 0.6, "test": 0.2, "validate": 0.2}


def compute_dataset_splits(
    df: pd.DataFrame,
    *,
    label_fracs: t.Optional[dict[str, float]] = None,
    shuffle: bool = True,
    seed: t.Optional[int] = None,
) -> pd.Series:
    """
    Split input dataset into random subsets with configurable proportions;
    by default, Databricks' standard train/test/validate splits are generated.

    Args:
        df
        label_fracs: Mapping of subset label to the (approximate) proportion of ``df``
            that gets split into it.
        shuffle: Whether or not to shuffle the data before splitting.
        seed: Optional integer used to set state for the underlying random generator;
            specify a value for reproducible splits, otherwise each call is unique.

    See Also:
        - :func:`sklearn.model_selection.train_test_split()`
    """
    if label_fracs is None:
        label_fracs = _DEFAULT_SPLIT_LABEL_FRACS

    labels = list(label_fracs.keys())
    fracs = list(label_fracs.values())
    rng = np.random.default_rng(seed=seed)
    return pd.Series(
        data=rng.choice(labels, size=len(df), p=fracs, shuffle=shuffle),
        index=df.index,
        dtype="string",
        name="split",
    )


def compute_sample_weights(
    df: pd.DataFrame,
    *,
    target_col: str = "target",
    class_weight: t.Literal["balanced"] | dict[object, int] = "balanced",
) -> pd.Series:
    """
    Estimate sample weights by class for imbalanced datasets.

    Args:
        df
        target_col: Name of column in ``df`` containing class label values
            i.e. "targets" to be predicted.
        class_weight: Weights associated with classes in the form ``{class_label: weight}``
            or "balanced" to automatically adjust weights inversely proportional to
            class frequencies in the input data.

    See Also:
        - :func:`sklearn.utils.class_weight.compute_sample_weight()`
    """
    return pd.Series(
        data=sklearn.utils.class_weight.compute_sample_weight(
            class_weight, df[target_col]
        ),
        index=df.index,
        dtype="float32",
        name="sample_weight",
    )



import functools as ft
import itertools
import logging
from collections.abc import Iterable
import typing as t

import pandas as pd
from pandas.core.groupby import DataFrameGroupBy

from .... import utils
from . import constants

LOGGER = logging.getLogger(__name__)


def add_features(
    df: pd.DataFrame, *, student_id_cols: list[str], sort_cols: list[str]
) -> pd.DataFrame:
    LOGGER.info("adding student-term cumulative features ...")
    # sort so that student-terms are ordered chronologically
    df = df.sort_values(by=student_id_cols + sort_cols, ignore_index=True)
    # specifically *don't* re-sort when grouping
    df_grped = df.groupby(by=student_id_cols, as_index=True, observed=True, sort=False)
    num_course_cols = [
        col
        for col in df.columns
        if col.startswith(f"{constants.NUM_COURSE_FEATURE_COL_PREFIX}_")
    ]
    dummy_course_cols = [
        col
        for col in df.columns
        if col.startswith(f"{constants.DUMMY_COURSE_FEATURE_COL_PREFIX}_")
    ]
    df_expanding_agg = (
        expanding_agg_features(
            df_grped,
            num_course_cols=num_course_cols,
            dummy_course_cols=dummy_course_cols,
            col_aggs=[
                ("term_id", "count"),
                ("term_in_peak_covid", "sum"),
                ("term_is_core", "sum"),
                ("term_is_noncore", "sum"),
                ("term_is_while_student_enrolled_at_other_inst", "sum"),
                ("term_is_pre_cohort", "sum"),
                ("course_level_mean", ["mean", "min", "std"]),
                ("course_grade_numeric_mean", ["mean", "min", "std"]),
                ("num_courses", ["sum", "mean", "min"]),
                ("num_credits_attempted", ["sum", "mean", "min"]),
                ("num_credits_earned", ["sum", "mean", "min"]),
                ("student_pass_rate_above_sections_avg", "sum"),
                ("student_completion_rate_above_sections_avg", "sum"),
            ],
            credits=constants.DEFAULT_COURSE_CREDIT_CHECK,
        )
        # rename/dtype special cols for clarity in downstream calcs
        .astype(
            {
                "cumcount_term_id": "Int8",
                "cumsum_term_is_core": "Int8",
                "cumsum_term_is_noncore": "Int8",
            }
        )
        .rename(
            columns={
                "cumcount_term_id": "cumnum_terms_enrolled",
                "cumsum_term_is_core": "cumnum_core_terms_enrolled",
                "cumsum_term_is_noncore": "cumnum_noncore_terms_enrolled",
            }
        )
    )
    df_cumnum_ur = cumnum_unique_and_repeated_features(
        df_grped, cols=["course_ids", "course_subjects", "course_subject_areas"]
    )
    concat_dfs = [df, df_cumnum_ur, df_expanding_agg]
    return (
        # despite best efforts, the student-id index is dropped from df_cumnum_ur
        # and, through sheer pandas insanity, merge on student_id_cols produces
        # huge numbers of duplicate rows -- truly impossible shit, that
        # however, by definition, these transforms shouldn't alter the original indexing, so:
        pd.concat(concat_dfs, axis="columns")
        # add a last couple features, which don't fit nicely into above logic
        .pipe(add_cumfrac_terms_enrolled_features, student_id_cols=student_id_cols)
        .pipe(
            add_term_diff_features,
            cols=[
                "num_courses",
                "num_credits_earned",
                "course_grade_numeric_mean",
                "course_level_mean",
            ],
            max_term_num=4,
            student_id_cols=student_id_cols,
        )
    )


def expanding_agg_features(
    df_grped: DataFrameGroupBy,
    *,
    num_course_cols: list[str],
    col_aggs: list[tuple[str, str | list[str]]],
    credits: t.Optional[int] = constants.DEFAULT_COURSE_CREDIT_CHECK,
    dummy_course_cols: t.Optional[list[str]] = None,
) -> pd.DataFrame:
    """
    Compute various aggregate features over an expanding window per (student) group.

    Args:
        df_grped
        num_course_cols
        col_aggs
        credits: the number of credits to check if courses of interest were taken within
        dummy_course_cols: the columns that were checked for whether a course was taken to check if they were taken within the number of credits of interest
    """
    LOGGER.info("computing expanding window aggregate features ...")
    agg_dict = dict(col_aggs) | {col: "sum" for col in num_course_cols}
    if dummy_course_cols is not None:
        agg_dict |= {col: "max" for col in dummy_course_cols}

    df_cumaggs = (
        df_grped.expanding()
        .agg(agg_dict)
        # pandas does weird stuff when indexing on windowed operations
        # this should get us back to student_id_cols only on the index
        .reset_index(level=-1, drop=True)
    )

    # unfortunately, expanding doesn't support the "named agg" syntax
    # so we have to (flatten and) rename the columns manually
    # df_cumaggs.columns = [f"{col}_cum{fn}" for col, fn in df_cumaggs.columns]
    df_cumaggs.columns = [f"cum{fn}_{col}" for col, fn in df_cumaggs.columns]
    # for all num_courses_*_cumsum columns, divide by num_courses_cumsum
    # to get equivalent cumfracs, whose relative values are easier for models to learn from
    num_courses_cumsum_cols = [
        col
        for col in df_cumaggs.columns
        if (
            col.startswith(f"cumsum_{constants.NUM_COURSE_FEATURE_COL_PREFIX}_")
            and col != "cumsum_num_courses"  # HACK
        )
    ]
    df_cumfracs = (
        df_cumaggs[num_courses_cumsum_cols]
        .div(df_cumaggs["cumsum_num_courses"], axis="index")
        .rename(columns=lambda col: col.replace("cumsum", "cumfrac"))
    )
    concat_dfs = [df_cumaggs, df_cumfracs]
    if dummy_course_cols is not None:
        action_cols = [f"cummax_{dummy_course}" for dummy_course in dummy_course_cols]
        action_status_df = pd.DataFrame(index=df_cumaggs.index)

        for col in action_cols:
            # within_col = f"{col}_in_{credits}_creds"
            within_col = col.replace("cummax_", f"cummax_in_{credits}_creds_")
            action_status_df[within_col] = (df_cumaggs[col].astype(bool)) & (
                df_cumaggs["cumsum_num_credits_earned"] <= credits
            )

        action_status_df = action_status_df.groupby(
            level=df_cumaggs.index.names
        ).transform("max")
        concat_dfs.append(action_status_df)

    return (
        # append cum-frac features to all cum-agg features
        pd.concat(concat_dfs, axis="columns")
        # *drop* the original cumsum columns, since the derived cumfracs are sufficient
        .drop(columns=num_courses_cumsum_cols)
        # drop our student-id(s) index, it just causes trouble
        .reset_index(drop=True)
    )


def cumnum_unique_and_repeated_features(
    df_grped: DataFrameGroupBy, *, cols: list[str]
) -> pd.DataFrame:
    """
    Compute the cumulative number of repeated elements within each (student) group
    for a set of columns whose values are per-row lists of elements.

    Args:
        df_grped
        cols
    """
    LOGGER.info("computing cumulative elements features ...")
    df_accumulated_lists = df_grped[cols].transform(_expand_elements)
    return df_accumulated_lists.assign(
        **{
            f"cumnum_unique_{col}": ft.partial(num_unique_elements, col=col)
            for col in cols
        }
        | {
            f"cumnum_repeated_{col}": ft.partial(num_repeated_elements, col=col)
            for col in cols
        }
    ).drop(columns=cols)


def num_unique_elements(df: pd.DataFrame, *, col: str) -> pd.Series:
    return df[col].map(lambda eles: len(set(eles)), na_action="ignore").astype("Int16")


def num_repeated_elements(df: pd.DataFrame, *, col: str) -> pd.Series:
    return (
        df[col]
        .map(lambda eles: len(eles) - len(set(eles)), na_action="ignore")
        .astype("Int16")
    )


def _expand_elements(x: list) -> list:
    return list(itertools.accumulate(x, func=_concat_elements))


def _concat_elements(*args: Iterable) -> list:
    return list(itertools.chain(*args))


def add_cumfrac_terms_enrolled_features(
    df: pd.DataFrame, *, student_id_cols: list[str]
) -> pd.DataFrame:
    LOGGER.info("computing cumfrac terms enrolled features ...")
    df_grped = df.groupby(by=student_id_cols, as_index=False, sort=False)
    df_min_term_ranks = df_grped.agg(
        min_student_term_rank=("term_rank", "min"),
        min_student_term_rank_core=("term_rank_core", "min"),
        min_student_term_rank_noncore=("term_rank_noncore", "min"),
    )
    return pd.merge(df, df_min_term_ranks, on=student_id_cols, how="inner").assign(
        cumfrac_terms_enrolled=ft.partial(
            _compute_cumfrac_terms_enrolled,
            term_rank_col="term_rank",
            min_student_term_rank_col="min_student_term_rank",
            cumnum_terms_enrolled_col="cumnum_terms_enrolled",
        ),
        cumfrac_core_terms_enrolled=ft.partial(
            _compute_cumfrac_terms_enrolled,
            term_rank_col="term_rank_core",
            min_student_term_rank_col="min_student_term_rank_core",
            cumnum_terms_enrolled_col="cumnum_core_terms_enrolled",
        ),
        cumfrac_noncore_terms_enrolled=ft.partial(
            _compute_cumfrac_terms_enrolled,
            term_rank_col="term_rank_noncore",
            min_student_term_rank_col="min_student_term_rank_noncore",
            cumnum_terms_enrolled_col="cumnum_noncore_terms_enrolled",
        ),
    )


def add_term_diff_features(
    df: pd.DataFrame,
    *,
    cols: list[str],
    max_term_num: int = 4,
    student_id_cols: str | list[str] = "student_id",
    term_num_col: str = "cumnum_terms_enrolled",
) -> pd.DataFrame:
    """
    Compute term-over-term differences per student for a set of columns, up to a specified
    maximum term number, where term numbers are _relative_, per student.

    Args:
        df
        cols
        max_term_num
        student_id_cols
        term_num_col
    """
    LOGGER.info("computing term diff features ...")
    student_id_cols = utils.types.to_list(student_id_cols)
    df_grped = (
        df.loc[:, student_id_cols + [term_num_col] + cols]
        .groupby(by=student_id_cols)
    )  # fmt: skip
    df_diffs = df.assign(
        **{f"{col}_diff_prev_term": df_grped[col].transform("diff") for col in cols}
    )
    df_pivots = []
    for col in cols:
        df_pivot_ = df_diffs.pivot(
            index=student_id_cols, columns=term_num_col, values=f"{col}_diff_prev_term"
        )
        # col 1 is always null (since no preceding periods to diff against)
        cols_to_drop = [1] + [
            col for col in df_pivot_.columns if int(col) > max_term_num
        ]
        df_pivots.append(
            df_pivot_.drop(columns=cols_to_drop)
            .rename(columns=lambda term_num: f"{col}_diff_term_{term_num - 1:.0f}_to_term_{term_num:.0f}")
        )  # fmt: skip
    df_pivot = pd.concat(df_pivots, axis="columns")
    return pd.merge(
        df_diffs, df_pivot, left_on=student_id_cols, right_index=True, how="left"
    )


def _compute_cumfrac_terms_enrolled(
    df: pd.DataFrame,
    *,
    term_rank_col: str = "term_rank",
    min_student_term_rank_col: str = "min_student_term_rank",
    cumnum_terms_enrolled_col: str = "cumnum_terms_enrolled",
) -> pd.Series:
    cumnum_terms_total = (df[term_rank_col] - df[min_student_term_rank_col]) + 1
    cumfrac_terms_enrolled = df[cumnum_terms_enrolled_col] / cumnum_terms_total
    return cumfrac_terms_enrolled.astype("Float32")


#######################
# DEPRECATED / BACKUP #
# only dig into these if databricks starts freaking out about expanding aggs func above

# def add_student_term_cumulative_features_databricks_hack(
#     df: pd.DataFrame,
#     *,
#     grp_cols: list[str],
#     sort_cols: list[str],
#     col_aggs: list[tuple[str, str | list[str]]],
# ) -> pd.DataFrame:
#     """
#     Compute features that aggregate cumulatively over each students' term history,
#     and assign them as additional columns.

#     Args:
#         df: Student-term features data, as output by :func:`make_student_term_dataset()` .
#         grp_cols: Columns needed to uniquely identify students,
#             used for grouping the rows in ``df`` .
#         sort_cols: Columns that sort rows within groups in chronological order,
#             such as term rank, term start dt, or "raw" (year, term) pairs.
#         col_aggs: Sequence of (column, agg-or-aggs) pairs, where each agg
#             is a pandas-compliant aggfunc.

#     Warning:
#         Under the hood, we've been forced to convert from standard pandas aggfuncs
#         into (a subset of) "transformation"-style cumulative methods. This is non-ideal,
#         but Databricks is behaving poorly -- and nondeterministically! -- when running
#         the unhacked sibling function above. So, in practice, the only supported aggs are
#         min, max, sum, and mean.

#     See Also:
#         - :func:`add_student_term_cumulative_features()`
#         - https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html#built-in-transformation-methods
#     """
#     LOGGER.info("adding student-term cumulative features ...")
#     df_grped = (
#         df.sort_values(by=grp_cols + sort_cols, ignore_index=True)
#         # specifically *don't* re-sort when grouping
#         .groupby(by=grp_cols, observed=True, sort=False)
#     )

#     cumagg_to_cols = {
#         "cumcount": [sort_cols[0]],
#         "cummin": _get_cols_for_agg(col_aggs, "min"),
#         "cummax": _get_cols_for_agg(col_aggs, "max"),
#         "cumsum": (
#             _get_cols_for_agg(col_aggs, "sum")
#             + [
#                 col
#                 for col in df.columns
#                 if col.startswith(constants.NUM_COURSE_FEATURE_COL_PREFIX)
#                 and col != "num_courses_enrolled"
#             ]
#         ),
#         "cummean": (
#             _get_cols_for_agg(col_aggs, "mean")
#             + [
#                 col
#                 for col in df.columns
#                 if col.startswith(constants.FRAC_COURSE_FEATURE_COL_PREFIX)
#             ]
#         ),
#     }
#     dfs = []
#     for cumagg, cols in cumagg_to_cols.items():
#         if cumagg == "cumcount":
#             df_ = (
#                 df_grped[cols[0]]
#                 .transform("cumcount")
#                 .to_frame(name="cumnum_student_terms")
#             )
#         elif cumagg == "cummean":
#             df_ = (
#                 df_grped[cols]
#                 .expanding()
#                 .mean()
#                 .rename(columns={col: f"{col}_{cumagg}" for col in cols})
#                 .reset_index(level=list(range(len(grp_cols))), drop=True)
#             )
#         else:
#             df_ = (
#                 df_grped[cols]
#                 .transform(cumagg)
#                 .rename(columns={col: f"{col}_{cumagg}" for col in cols})
#             )
#         dfs.append(df_)

#     # make sure each cum-agg'd df has the same number of rows
#     assert len(set(len(df_) for df_ in dfs)) == 1

#     df_cfs = (
#         pd.concat(dfs, axis="columns")
#         .rename(
#             columns={
#                 "course_id_nunique_cummean": "nunique_courses_cummean",
#                 "course_cip_nunique_cummean": "nunique_course_subjects_cummean",
#             },
#         )
#         # let's also sneak in a derivative cumulative feature
#         .assign(
#             num_credits_earned_cumfrac=lambda df: (
#                 df["num_credits_earned_cumsum"] / df["num_credits_attempted_cumsum"]
#             )
#         )
#     )
#     return pd.concat([df, df_cfs], axis="columns")


# def _get_cols_for_agg(
#     col_aggs: list[tuple[str, str | list[str]]], agg: str
# ) -> list[str]:
#     return [
#         col_
#         for col_, agg_ in col_aggs
#         if (isinstance(agg_, str) and agg_ == agg) or agg in agg_
#     ]


# def compute_cumfrac_terms_unenrolled(
#     df: pd.DataFrame, *, grp_cols: list[str], term_rank_col: str
# ) -> pd.Series:
#     df_grped = df.groupby(by=grp_cols)
#     cumnum_terms_enrolled = df_grped[term_rank_col].cumcount() + 1
#     cumnum_terms_unenrolled = (
#         df_grped[term_rank_col]
#         .transform(lambda s: s.diff().sub(1.0).cumsum())
#         .fillna(0.0)
#         .astype("int64")
#     )
#     cumnum_terms_total = cumnum_terms_enrolled + cumnum_terms_unenrolled
#     cumfrac_terms_unenrolled = cumnum_terms_unenrolled.div(cumnum_terms_total)
#     return cumfrac_terms_unenrolled



"""Task for validating data using TFDV.

This task can be used to validate data against a reference schema, generate a
set of statistics from a data source and raise an exception if anomalies are
detected. It can be used to validate data in both training and inference
pipelines.
"""

import argparse
import logging
import os

from databricks.connect import DatabricksSession
import pandas as pd
import pyspark
import tensorflow_data_validation as tfdv
from tensorflow_metadata.proto.v0 import anomalies_pb2
from tensorflow_metadata.proto.v0 import schema_pb2
from tensorflow_metadata.proto.v0 import statistics_pb2


# pylint: disable=no-member
class DataValidationTask:
    """Task for validating data using TFDV."""

    def load_reference_schema(self, schema_path: str) -> schema_pb2.Schema:
        """Loads a reference schema from a pbtxt file.

        Args:
          schema_path: The path to the schema pbtxt file.

        Returns:
          The schema loaded from the pbtxt file.
        """
        logging.info("Loading reference schema from: %s", schema_path)
        return tfdv.load_schema_text(schema_path)

    def generate_statistics_from_delta_table(
        self,
        spark_session: pyspark.sql.session.SparkSession,
        input_table_path: str,
        table_format: str = "delta",
    ) -> statistics_pb2.DatasetFeatureStatisticsList:
        """Generates statistics from a delta table.

        Args:
          spark_session: The Spark session to use for reading the table.
          input_table_path: The path to the delta table to read.
          table_format: The format of the delta table.

        Returns:
          The statistics generated from the delta table.
        """
        logging.info("Generating statistics from delta table: %s", input_table_path)
        df = spark_session.read.format(table_format).table(input_table_path).toPandas()
        logging.info("Generated statistics from delta table.")
        return tfdv.generate_statistics_from_dataframe(df)

    def generate_statistics_from_csv(
        self, input_csv_path: str
    ) -> statistics_pb2.DatasetFeatureStatisticsList:
        """Generates statistics from a CSV file.

        Args:
          input_csv_path: The path to the CSV file to read.

        Returns:
          The statistics generated from the CSV file.
        """
        logging.info("Generating statistics from csv: %s", input_csv_path)
        df = pd.read_csv(input_csv_path)
        return tfdv.generate_statistics_from_dataframe(df)

    def generate_anomalies(
        self,
        stats: statistics_pb2.DatasetFeatureStatisticsList,
        schema: schema_pb2.Schema,
        environment: str,
    ) -> anomalies_pb2.Anomalies:
        """Generates anomalies from the statistics and schema.

        Args:
          stats: The statistics to validate.
          schema: The schema to validate against.
          environment: The environment to validate in (TRAINING or SERVING).

        Returns:
          The anomalies generated from the statistics and schema.
        """
        logging.info("Generating anomalies.")
        anomalies = tfdv.validate_statistics(
            statistics=stats, schema=schema, environment=environment
        )
        return anomalies

    def save_anomalies(
        self, anomalies: anomalies_pb2.Anomalies, output_path: str
    ) -> None:
        """Saves the anomalies to a pbtxt file.

        Args:
          anomalies: The anomalies to save.
          output_path: The path to save the anomalies to.

        Returns:
          None
        """
        logging.info("Saving anomalies to: %s", output_path)
        tfdv.write_anomalies_text(
            anomalies, os.path.join(output_path, "anomalies.pbtxt")
        )

    def save_statistics(
        self, statistics: statistics_pb2.DatasetFeatureStatisticsList, output_path: str
    ) -> None:
        """Saves the statistics to a pbtxt file.

        Args:
          statistics: The summary statistics to save.
          output_path: The path to save the statistics to.

        Returns:
          None
        """
        logging.info("Saving statistics to: %s", output_path)
        tfdv.write_stats_text(statistics, os.path.join(output_path, "statistics.pbtxt"))

    def check_for_anomalies(
        self,
        anomalies: anomalies_pb2.Anomalies,
        output_path: str,
        fail_on_anomalies: bool = False,
    ) -> None:
        """Checks for anomalies in the anomalies proto.

        Args:
          anomalies: The anomalies to check.
          output_path: The path to save the anomalies to.
          fail_on_anomalies: Whether to fail the job if anomalies are found.

        Returns:
          None
        """
        self.save_anomalies(anomalies, output_path)
        logging.info(
            "Checking for anomalies. fail_on_anomalies is set to %s", fail_on_anomalies
        )
        if anomalies.anomaly_info:
            logging.warning("Anomalies found. Check: %s/anomalies.pbtxt", output_path)
            if fail_on_anomalies:
                logging.error(anomalies.anomaly_info.items())
                raise ValueError("Anomalies found, pipline error.")
            logging.warning(anomalies.anomaly_info.items())
        else:
            logging.info("No anomalies found.")


def main():
    """Main function for validating data using TFDV."""
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input_table_path", required=True, help="Path to input table to validate."
    )
    parser.add_argument(
        "--input_table_format",
        required=False,
        default="delta",
        help="Input data table format.",
    )
    parser.add_argument(
        "--input_schema_path", required=True, help="Path to input schema."
    )
    parser.add_argument(
        "--output_artifact_path", required=True, help="Path to write output artifacts."
    )
    parser.add_argument(
        "--environment",
        required=True,
        help="Environment to use for validation (TRAINING or SERVING).",
    )
    parser.add_mutually_exclusive_group(required=False)
    parser.add_argument(
        "--fail_on_anomalies_true", dest="fail_on_anomalies", action="store_true"
    )
    parser.add_argument(
        "--fail_on_anomalies_false", dest="fail_on_anomalies", action="store_false"
    )
    parser.set_defaults(fail_on_anomalies=True)
    args = parser.parse_args()

    task = DataValidationTask()

    spark_session = DatabricksSession.builder.getOrCreate()
    stats = task.generate_statistics_from_delta_table(
        spark_session, args.input_table_path, args.input_table_format
    )
    task.save_statistics(stats, args.output_artifact_path)
    schema = task.load_reference_schema(args.input_schema_path)
    anomalies = task.generate_anomalies(stats, schema, args.environment)
    task.check_for_anomalies(
        anomalies, args.output_artifact_path, fail_on_anomalies=args.fail_on_anomalies
    )


if __name__ == "__main__":
    main()



# Databricks notebook source
# MAGIC %md
# MAGIC # SST Make and Explain Predictions
# MAGIC
# MAGIC Fourth step in the process of transforming rawdata into actionable, data-driven insights for advisors: generate predictions and feature importances for new (unlabeled) data.
# MAGIC
# MAGIC ### References
# MAGIC
# MAGIC - [Data science product components (Confluence doc)](https://datakind.atlassian.net/wiki/spaces/TT/pages/237862913/Data+science+product+components+the+modeling+process)
# MAGIC - [Databricks runtimes release notes](https://docs.databricks.com/en/release-notes/runtime/index.html)

# COMMAND ----------

# MAGIC %md
# MAGIC # setup

# COMMAND ----------

# MAGIC %sh python --version

# COMMAND ----------

# WARNING: AutoML/mlflow expect particular packages with version constraints
# that directly conflicts with dependencies in our SST repo. As a temporary fix,
# we need to manually install a certain version of pandas and scikit-learn in order
# for our models to load and run properly.

# %pip install "student-success-tool==0.3.7"
# %pip install "pandas==1.5.3"
# %pip install "scikit-learn==1.3.0"
# %restart_python

# COMMAND ----------

import functools as ft
import logging

import mlflow
import pandas as pd
import shap
from databricks.connect import DatabricksSession
from pyspark.sql.types import FloatType, StringType, StructField, StructType

from student_success_tool import configs, dataio, modeling
from student_success_tool.modeling import inference, evaluation

# COMMAND ----------

logging.basicConfig(level=logging.INFO, force=True)
logging.getLogger("py4j").setLevel(logging.WARNING)  # ignore databricks logger

try:
    spark = DatabricksSession.builder.getOrCreate()
except Exception:
    logging.warning("unable to create spark session; are you in a Databricks runtime?")
    pass

try:
    # Get the pipeline type from job definition
    run_type = dbutils.widgets.get("run_type")  # noqa: F821
except Py4JJavaError:
    # Run script interactively
    run_type = "predict"

# COMMAND ----------

# Databricks logs every instance that uses sklearn or other modelling libraries
# to MLFlow experiments... which we don't want
mlflow.autolog(disable=True)

# COMMAND ----------

# MAGIC %md
# MAGIC ## import school-specific code

# COMMAND ----------

# project configuration stored as a config file in TOML format
cfg = dataio.read_config(
    "./config-TEMPLATE.toml", schema=configs.custom.CustomProjectConfig
)
cfg

# COMMAND ----------

# Assign student ID column name
student_id_col = cfg.student_id_col

# Load custom human-friendly feature names
features_table = dataio.read_features_table("./features_table.toml")

# COMMAND ----------

# MAGIC %md
# MAGIC # load artifacts

# COMMAND ----------

df = dataio.read.from_delta_table(
    cfg.datasets.silver["modeling"].table_path,
    spark_session=spark,
)
df.head()

# COMMAND ----------

model = dataio.models.load_mlflow_model(
    cfg.model.mlflow_model_uri,
    cfg.model.framework,
)
model

# COMMAND ----------

model_feature_names = model.named_steps["column_selector"].get_params()["cols"]
logging.info(
    "model uses %s features: %s", len(model_feature_names), model_feature_names
)

# COMMAND ----------

df_train = modeling.evaluation.extract_training_data_from_model(cfg.model.experiment_id)
if cfg.split_col:
    df_train = df_train.loc[df_train[cfg.split_col].eq("train"), :]
df_train.shape

# COMMAND ----------

# MAGIC %md
# MAGIC # make predictions

# COMMAND ----------

if cfg.split_col and cfg.split_col in df.columns:
    df_test = df.loc[df[cfg.split_col].eq("test"), :]
else:
    df_test = df.copy(deep=True)

if cfg.target_col in df.columns:
    df_test = df_test.drop(columns=cfg.target_col)
    # TODO: do we want to sample this dataset for labeled data?
    df_test = df_test.sample(n=200)

# COMMAND ----------

features = df_test.loc[:, model_feature_names]
unique_ids = df_test[student_id_col]

# COMMAND ----------

pred_probs = modeling.inference.predict_probs(
    features,
    model=model,
    feature_names=model_feature_names,
    pos_label=cfg.pos_label,
)
print(pred_probs.shape)
pd.Series(pred_probs).describe()

# COMMAND ----------

# MAGIC %md
# MAGIC # explain predictions

# COMMAND ----------

shap_ref_data_size = 100
# SHAP can't explain models using data with nulls
# so, impute nulls using the mode (most frequent values)
train_mode = df_train.mode().iloc[0]
# sample training dataset as "reference" data for SHAP Explainer
df_ref = (
    df_train.sample(
        n=min(shap_ref_data_size, df_train.shape[0]),
        random_state=cfg.random_state,
    )
    .fillna(train_mode)
    .loc[:, model_feature_names]
)
df_ref.shape

# COMMAND ----------

explainer = shap.explainers.KernelExplainer(
    ft.partial(
        modeling.inference.predict_probs,
        model=model,
        feature_names=model_feature_names,
        pos_label=cfg.pos_label,
    ),
    df_ref,
    link="identity",
)
explainer

# COMMAND ----------

shap_schema = StructType(
    [StructField(student_id_col, StringType(), nullable=False)]
    + [StructField(col, FloatType(), nullable=False) for col in model_feature_names]
)

df_shap_values = (
    spark.createDataFrame(
        df_test.reindex(columns=model_feature_names + [student_id_col])
    )  # noqa: F821
    .repartition(sc.defaultParallelism)  # noqa: F821
    .mapInPandas(
        ft.partial(
            inference.calculate_shap_values_spark_udf,
            student_id_col=student_id_col,
            model_features=model_feature_names,
            explainer=explainer,
            mode=train_mode,
        ),
        schema=shap_schema,
    )
    .toPandas()
    .set_index(student_id_col)
    .reindex(df_test[student_id_col])
    .reset_index(drop=False)
)
df_shap_values

# COMMAND ----------

with mlflow.start_run(run_id=cfg.model.run_id):
    # Create & log SHAP summary plot
    inference.shap_summary_plot(
        df_shap_values=df_shap_values,
        df_test=df_test,
        model_feature_names=model_feature_names,
        model_classes=model.classes_,
    )

    # Create & log ranked features by SHAP magnitude
    selected_features_df = inference.generate_ranked_feature_table(
        features,
        df_shap_values[model_feature_names].to_numpy(),
        features_table=features_table,
    )

selected_features_df

# COMMAND ----------

# MAGIC %md
# MAGIC # finalize results

# COMMAND ----------

# Provide output using top features, SHAP values, and support scores
result = inference.select_top_features_for_display(
    features,
    unique_ids,
    pred_probs,
    df_shap_values[model_feature_names].to_numpy(),
    n_features=cfg.inference.num_top_features,
    features_table=features_table,
    needs_support_threshold_prob=cfg.inference.min_prob_pos_label,
)
result

# COMMAND ----------

# save sample advisor output dataset
dataio.write.to_delta_table(
    result, cfg.datasets.gold["advisor_output"].table_path, spark_session=spark
)

# COMMAND ----------

# MAGIC %md
# MAGIC ## Log Front-End Tables

# COMMAND ----------

# Log MLFlow confusion matrix & roc table figures in silver schema

with mlflow.start_run() as run:
    confusion_matrix = evaluation.log_confusion_matrix(
        institution_id=cfg.institution_id,
        automl_run_id=cfg.model.run_id,
    )

    # Log roc curve table for front-end
    roc_logs = evaluation.log_roc_table(
        institution_id=cfg.institution_id,
        automl_run_id=cfg.model.run_id,
        modeling_dataset_name=cfg.datasets.silver.modeling.table_path,
    )

# COMMAND ----------

shap_feature_importance = inference.generate_ranked_feature_table(
    features, df_shap_values[model_feature_names].to_numpy(), features_table
)
shap_feature_importance


# COMMAND ----------

# save sample advisor output dataset
dataio.write.to_delta_table(
    shap_feature_importance,
    f"staging_sst_01.{cfg.institution_id}_silver.training_{cfg.model.run_id}_shap_feature_importance",
    spark_session=spark,
)

# COMMAND ----------

support_score_distribution = inference.support_score_distribution_table(
    features,
    unique_ids,
    pred_probs,
    df_shap_values[model_feature_names],
    cfg.inference.dict(),
    features_table,
    model_feature_names,
)
support_score_distribution

# COMMAND ----------

# save sample advisor output dataset
dataio.write.to_delta_table(
    support_score_distribution,
    f"staging_sst_01.{cfg.institution_id}_silver.training_{cfg.model.run_id}_support_overview",
    spark_session=spark,
)



import os
import sys
import json
import shutil
from datetime import datetime
from typing import Any, Dict, List, Optional


class SimpleLogger:
    """
    A JSONL logger that temporarily moves the institution log file to /tmp,
    appends the new entry, and moves it back to the source directory.
    """

    def __init__(self, log_path: str, institution_id: Optional[str] = None):
        self._institution_id = institution_id
        self._final_log_path = log_path

        tmp_dir = "/tmp/logs"
        os.makedirs(tmp_dir, exist_ok=True)

        # Use a temp file per institution to avoid collision
        self._tmp_log_path = os.path.join(
            tmp_dir, f"{institution_id}_validation.tmp.log"
        )

        # If final log exists, move it into tmp before writing
        if os.path.exists(self._final_log_path):
            shutil.copy2(self._final_log_path, self._tmp_log_path)
        else:
            # Create empty temp log file
            open(self._tmp_log_path, "w", encoding="utf-8").close()

        # Open for appending
        self._fh = open(self._tmp_log_path, "a", encoding="utf-8")

    def _write(self, entry: Dict[str, Any]) -> None:
        entry.setdefault("timestamp", datetime.utcnow().isoformat() + "Z")
        line = json.dumps(entry, default=str, indent=4)
        self._fh.write(line + "\n")
        self._fh.flush()
        sys.stdout.write(line + "\n")
        sys.stdout.flush()

    def error(
        self,
        message: Optional[str] = None,
        *,
        extra_columns: Optional[List[str]] = None,
        missing_required: Optional[List[str]] = None,
        schema_errors: Any = None,
        failure_cases: Any = None,
    ) -> None:
        entry: Dict[str, Any] = {"validation_status": "hard_error"}
        if message:
            entry["message"] = message
        if extra_columns is not None:
            entry["extra_columns"] = extra_columns
        if missing_required is not None:
            entry["missing_required"] = missing_required
        if schema_errors is not None:
            entry["schema_errors"] = schema_errors
        if failure_cases is not None:
            entry["failure_cases"] = failure_cases
        self._write(entry)

    def info(self, *, missing_optional: Optional[List[str]] = None) -> None:
        status = "passed_with_soft_errors" if missing_optional else "passed"
        entry = {
            "validation_status": status,
            "missing_optional": missing_optional or [],
        }
        self._write(entry)

    def exception(self, message: str = "Unexpected exception occurred") -> None:
        exc_type, exc_val, _ = sys.exc_info()
        entry = {
            "validation_status": "exception",
            "message": message,
            "error_type": exc_type.__name__ if exc_type else None,
            "error": str(exc_val),
        }
        self._write(entry)

    def close(self) -> None:
        try:
            self._fh.close()
            final_dir = os.path.dirname(self._final_log_path)
            os.makedirs(final_dir, exist_ok=True)
            shutil.move(self._tmp_log_path, self._final_log_path)
        except Exception as e:
            sys.stderr.write(f"[LOGGER] Failed to move final log file: {e}\n")


def setup_logger(
    institution_id: Optional[str] = None, log_file: str = "validation.log"
) -> SimpleLogger:
    if not institution_id:
        raise ValueError("institution_id is required for institution-specific logging")

    log_dir = f"/Volumes/staging_sst_01/{institution_id}_bronze/bronze_volume/logs"
    log_path = os.path.join(log_dir, log_file)

    return SimpleLogger(log_path=log_path, institution_id=institution_id)



import re
import typing as t
from collections.abc import Iterable

from . import types

RE_VARIOUS_PUNCTS = re.compile(r"[!()*+\,\-./:;<=>?[\]^_{|}~]")
RE_QUOTATION_MARKS = re.compile(r"[\'\"\`]")


def unique_elements_in_order(eles: Iterable) -> Iterable:
    """Get unique elements from an iterable, in order of appearance."""
    seen = set()  # type: ignore
    seen_add = seen.add
    for ele in eles:
        if ele not in seen:
            seen_add(ele)
            yield ele


def convert_to_snake_case(col: str) -> str:
    """Convert column name into snake case, without punctuation."""
    col = RE_VARIOUS_PUNCTS.sub(" ", col)
    col = RE_QUOTATION_MARKS.sub("", col)
    # TODO: *pretty sure* this could be cleaner and more performant, but shrug
    words = re.sub(
        r"([A-Z][a-z]+)", r" \1", re.sub(r"([A-Z]+|[0-9]+|\W+)", r" \1", col)
    ).split()
    return "_".join(words).lower()


def convert_intensity_time_limits(
    unit: t.Literal["term", "year"],
    intensity_time_limits: types.IntensityTimeLimitsType,
    *,
    num_terms_in_year: int,
) -> dict[str, float]:
    """
    Convert enrollment intensity-specific time limits into a particular ``unit`` ,
    whether input limits were given in units of years or terms.

    Args:
        unit: The time unit into which inputs are converted, either "term" or "year".
        intensity_time_limits: Mapping of enrollment intensity value (e.g. "FULL-TIME")
            to the maximum number of years or terms (e.g. [4.0, "year"], [12.0, "term"])
            considered "success" for a school in their particular use case.
        num_terms_in_year: Number of academic terms in one academic year,
            used to convert between term- and year-based time limits;
            for example: 4 => FALL, WINTER, SPRING, and SUMMER terms.
    """
    if unit == "year":
        intensity_nums = {
            intensity: num if unit == "year" else num / num_terms_in_year
            for intensity, (num, unit) in intensity_time_limits.items()
        }
    else:
        intensity_nums = {
            intensity: num if unit == "term" else num * num_terms_in_year
            for intensity, (num, unit) in intensity_time_limits.items()
        }
    return intensity_nums
